# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.fab.auth_manager.security_manager.override import FabAirflowSecurityManagerOverride

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


class FakeAppBuilder:
    """
    Stand-in class to replace a Flask App Builder.

    The only purpose is to provide the ``self.appbuilder.get_session`` interface
    for ``ApplessAirflowSecurityManager`` so it can be used without a real Flask
    app, which is slow to create.
    """

    def __init__(self, session: Session | None = None) -> None:
        self.session = session


class ApplessAirflowSecurityManager(FabAirflowSecurityManagerOverride):
    """Security Manager that doesn't need the whole flask app."""

    def __init__(self, session: Session | None = None):
        self.appbuilder = FakeAppBuilder(session)
        self._session = session

    @property
    def session(self):
        return self._session



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# mypy: disable-error-code=var-annotated
from __future__ import annotations

import logging
from functools import reduce
from typing import TYPE_CHECKING

from flask import Blueprint, current_app, url_for
from flask_appbuilder import __version__
from flask_appbuilder.babel.manager import BabelManager
from flask_appbuilder.const import (
    LOGMSG_ERR_FAB_ADD_PERMISSION_MENU,
    LOGMSG_ERR_FAB_ADD_PERMISSION_VIEW,
    LOGMSG_ERR_FAB_ADDON_IMPORT,
    LOGMSG_ERR_FAB_ADDON_PROCESS,
    LOGMSG_INF_FAB_ADD_VIEW,
    LOGMSG_INF_FAB_ADDON_ADDED,
    LOGMSG_WAR_FAB_VIEW_EXISTS,
)
from flask_appbuilder.filters import TemplateFilters
from flask_appbuilder.menu import Menu
from flask_appbuilder.views import IndexView, UtilView

from airflow import settings
from airflow.api_fastapi.app import create_auth_manager, get_auth_manager
from airflow.configuration import conf
from airflow.providers.fab.www.security_manager import AirflowSecurityManagerV2
from airflow.providers.fab.www.views import FabIndexView, redirect

if TYPE_CHECKING:
    from flask import Flask
    from flask_appbuilder import BaseView
    from flask_appbuilder.security.manager import BaseSecurityManager
    from sqlalchemy.orm import Session


# This product contains a modified portion of 'Flask App Builder' developed by Daniel Vaz Gaspar.
# (https://github.com/dpgaspar/Flask-AppBuilder).
# Copyright 2013, Daniel Vaz Gaspar
# This module contains code imported from FlaskAppbuilder, so lets use _its_ logger name
log = logging.getLogger("flask_appbuilder.base")


def dynamic_class_import(class_path):
    """
    Will dynamically import a class from a string path.

    :param class_path: string with class path
    :return: class
    """
    # Split first occurrence of path
    try:
        tmp = class_path.split(".")
        module_path = ".".join(tmp[0:-1])
        package = __import__(module_path)
        return reduce(getattr, tmp[1:], package)
    except Exception as e:
        log.exception(e)
        log.error(LOGMSG_ERR_FAB_ADDON_IMPORT, class_path, e)


class AirflowAppBuilder:
    """This is the base class for all the framework."""

    baseviews: list[BaseView | Session] = []
    # Database Session
    session = None
    # Security Manager Class
    sm: BaseSecurityManager
    # Babel Manager Class
    bm = None
    # dict with addon name has key and instantiated class has value
    addon_managers: dict
    # temporary list that hold addon_managers config key
    _addon_managers: list

    menu = None
    indexview = None

    static_folder = None
    static_url_path = None

    template_filters = None

    def __init__(
        self,
        app=None,
        session: Session | None = None,
        menu=None,
        indexview=None,
        base_template="airflow/main.html",
        static_folder="static/appbuilder",
        static_url_path="/appbuilder",
        enable_plugins: bool = False,
    ):
        """
        App-builder constructor.

        :param app:
            The flask app object
        :param session:
            The SQLAlchemy session object
        :param menu:
            optional, a previous constructed menu
        :param indexview:
            optional, your customized indexview
        :param static_folder:
            optional, your override for the global static folder
        :param static_url_path:
            optional, your override for the global static url path
        :param enable_plugins:
            optional, whether plugins are enabled for this app. AirflowAppBuilder from FAB provider can be
            instantiated in two modes:
             - Plugins enabled. The Flask application is responsible to execute Airflow 2 plugins.
               This application is only running if there are Airflow 2 plugins defined as part of the Airflow
               environment
             - Plugins disabled. The Flask application is responsible to execute the FAB auth manager login
               process. This application is only running if FAB auth manager is the auth manager configured
               in the Airflow environment
        """
        from airflow.providers_manager import ProvidersManager

        providers_manager = ProvidersManager()
        providers_manager.initialize_providers_configuration()
        self.baseviews = []
        self._addon_managers = []
        self.addon_managers = {}
        self.menu = menu
        self.base_template = base_template
        self.indexview = indexview
        self.static_folder = static_folder
        self.static_url_path = static_url_path
        self.enable_plugins = enable_plugins
        self.update_perms = conf.getboolean("fab", "UPDATE_FAB_PERMS")
        self.auth_rate_limited = conf.getboolean("fab", "AUTH_RATE_LIMITED")
        self.auth_rate_limit = conf.get("fab", "AUTH_RATE_LIMIT")
        if app is not None:
            self.init_app(app, session)

    def init_app(self, app, session):
        """
        Will initialize the Flask app, supporting the app factory pattern.

        :param app:
        :param session: The SQLAlchemy session
        """
        log.info("Initializing AppBuilder")
        app.config.setdefault("APP_NAME", "F.A.B.")
        app.config.setdefault("APP_THEME", "")
        app.config.setdefault("APP_ICON", "")
        app.config.setdefault("LANGUAGES", {"en": {"flag": "gb", "name": "English"}})
        app.config.setdefault("ADDON_MANAGERS", [])
        app.config.setdefault("RATELIMIT_ENABLED", self.auth_rate_limited)
        app.config.setdefault("FAB_BASE_TEMPLATE", self.base_template)
        app.config.setdefault("FAB_STATIC_FOLDER", self.static_folder)
        app.config.setdefault("FAB_STATIC_URL_PATH", self.static_url_path)
        app.config.setdefault("AUTH_RATE_LIMITED", self.auth_rate_limited)
        app.config.setdefault("AUTH_RATE_LIMIT", self.auth_rate_limit)

        self.base_template = app.config.get("FAB_BASE_TEMPLATE", self.base_template)
        self.static_folder = app.config.get("FAB_STATIC_FOLDER", self.static_folder)
        self.static_url_path = app.config.get("FAB_STATIC_URL_PATH", self.static_url_path)
        _index_view = app.config.get("FAB_INDEX_VIEW", None)
        if _index_view is not None:
            self.indexview = dynamic_class_import(_index_view)
        elif not self.enable_plugins:
            self.indexview = FabIndexView
        else:
            self.indexview = IndexView
        _menu = app.config.get("FAB_MENU", None)
        if _menu is not None:
            self.menu = dynamic_class_import(_menu)
        else:
            self.menu = self.menu or Menu()

        self._addon_managers = app.config["ADDON_MANAGERS"]
        self.session = session
        auth_manager = create_auth_manager()
        auth_manager.appbuilder = self
        if hasattr(auth_manager, "init_flask_resources"):
            auth_manager.init_flask_resources()
        if hasattr(auth_manager, "security_manager"):
            self.sm = auth_manager.security_manager
        else:
            self.sm = AirflowSecurityManagerV2(self)
        self.bm = BabelManager(self)
        self._add_global_static()
        self._add_global_filters()
        app.before_request(self.sm.before_request)
        self._add_admin_views()
        self._add_addon_views()
        self._init_extension(app)
        self._swap_url_filter()

    def _swap_url_filter(self):
        """Use our url filtering util function so there is consistency between FAB and Airflow routes."""
        from flask_appbuilder.security import views as fab_sec_views

        from airflow.providers.fab.www.views import get_safe_url

        fab_sec_views.get_safe_redirect = get_safe_url
        fab_sec_views.redirect = redirect

    def _init_extension(self, app):
        app.appbuilder = self
        if not hasattr(app, "extensions"):
            app.extensions = {}
        app.extensions["appbuilder"] = self

    @property
    def app(self) -> Flask:
        log.warning(
            "appbuilder.app is deprecated and will be removed in a future version. Use current_app instead"
        )
        return current_app

    @property
    def get_app(self) -> Flask:
        log.warning(
            "appbuilder.get_app is deprecated and will be removed in a future version. "
            "Use current_app instead"
        )
        return self.app

    @property
    def app_name(self):
        """
        Get the App name.

        :return: String with app name
        """
        return current_app.config["APP_NAME"]

    @property
    def app_theme(self):
        """
        Get the App theme name.

        :return: String app theme name
        """
        return current_app.config["APP_THEME"]

    @property
    def app_icon(self):
        """
        Get the App icon location.

        :return: String with relative app icon location
        """
        return current_app.config["APP_ICON"]

    @property
    def languages(self):
        return current_app.config["LANGUAGES"]

    @property
    def version(self):
        """
        Get the current F.A.B. version.

        :return: String with the current F.A.B. version
        """
        return __version__

    def _add_global_filters(self):
        self.template_filters = TemplateFilters(current_app, self.sm)

    def _add_global_static(self):
        bp = Blueprint(
            "appbuilder",
            "flask_appbuilder.base",
            url_prefix="/static",
            template_folder="templates",
            static_folder=self.static_folder,
            static_url_path=self.static_url_path,
        )
        current_app.register_blueprint(bp)

    def _add_admin_views(self):
        """Register indexview, utilview (back function), babel views and Security views."""
        self.indexview = self._check_and_init(self.indexview)
        self.add_view_no_menu(self.indexview)
        self.add_view_no_menu(UtilView())
        auth_manager = get_auth_manager()
        if hasattr(auth_manager, "register_views"):
            auth_manager.register_views()

    def _add_addon_views(self):
        """Register declared addons."""
        for addon in self._addon_managers:
            addon_class = dynamic_class_import(addon)
            if addon_class:
                # Instantiate manager with appbuilder (self)
                addon_class = addon_class(self)
                try:
                    addon_class.pre_process()
                    addon_class.register_views()
                    addon_class.post_process()
                    self.addon_managers[addon] = addon_class
                    log.info(LOGMSG_INF_FAB_ADDON_ADDED, addon)
                except Exception as e:
                    log.exception(e)
                    log.error(LOGMSG_ERR_FAB_ADDON_PROCESS, addon, e)

    def _check_and_init(self, baseview):
        if callable(baseview):
            baseview = baseview()
        return baseview

    def add_view(
        self,
        baseview,
        name,
        href="",
        icon="",
        label="",
        category="",
        category_icon="",
        category_label="",
        menu_cond=None,
    ):
        """
        Add your views associated with menus using this method.

        :param baseview:
            A BaseView type class instantiated or not.
            This method will instantiate the class for you if needed.
        :param name:
            The string name that identifies the menu.
        :param href:
            Override the generated link for the menu.
            You can use an url string or an endpoint name
            if non provided default_view from view will be set as link.
        :param icon:
            Font-Awesome icon name, optional.
        :param label:
            The label that will be displayed on the menu,
            if absent param name will be used
        :param category:
            The menu category where the menu will be included,
            if non provided the view will be accessible as a top menu.
        :param category_icon:
            Font-Awesome icon name for the category, optional.
        :param category_label:
            The label that will be displayed on the menu,
            if absent param name will be used
        :param menu_cond:
            If a callable, :code:`menu_cond` will be invoked when
            constructing the menu items. If it returns :code:`True`,
            then this link will be a part of the menu. Otherwise, it
            will not be included in the menu items. Defaults to
            :code:`None`, meaning the item will always be present.

        Examples::

            appbuilder = AppBuilder(app, db)
            # Register a view, rendering a top menu without icon.
            appbuilder.add_view(MyModelView(), "My View")
            # or not instantiated
            appbuilder.add_view(MyModelView, "My View")
            # Register a view, a submenu "Other View" from "Other" with a phone icon.
            appbuilder.add_view(MyOtherModelView, "Other View", icon="fa-phone", category="Others")
            # Register a view, with category icon and translation.
            appbuilder.add_view(
                YetOtherModelView,
                "Other View",
                icon="fa-phone",
                label=_("Other View"),
                category="Others",
                category_icon="fa-envelop",
                category_label=_("Other View"),
            )
            # Register a view whose menu item will be conditionally displayed
            appbuilder.add_view(
                YourFeatureView,
                "Your Feature",
                icon="fa-feature",
                label=_("Your Feature"),
                menu_cond=lambda: is_feature_enabled("your-feature"),
            )
            # Add a link
            appbuilder.add_link("google", href="www.google.com", icon="fa-google-plus")
        """
        baseview = self._check_and_init(baseview)
        log.debug(LOGMSG_INF_FAB_ADD_VIEW, baseview.__class__.__name__, name)

        if not self._view_exists(baseview):
            baseview.appbuilder = self
            self.baseviews.append(baseview)
            self._process_inner_views()
            self.register_blueprint(baseview)
            self._add_permission(baseview)
            self.add_limits(baseview)
        self.add_link(
            name=name,
            href=href,
            icon=icon,
            label=label,
            category=category,
            category_icon=category_icon,
            category_label=category_label,
            baseview=baseview,
            cond=menu_cond,
        )
        return baseview

    def add_link(
        self,
        name,
        href,
        icon="",
        label="",
        category="",
        category_icon="",
        category_label="",
        baseview=None,
        cond=None,
    ):
        """
        Add your own links to menu using this method.

        :param name:
            The string name that identifies the menu.
        :param href:
            Override the generated link for the menu.
            You can use an url string or an endpoint name
        :param icon:
            Font-Awesome icon name, optional.
        :param label:
            The label that will be displayed on the menu,
            if absent param name will be used
        :param category:
            The menu category where the menu will be included,
            if non provided the view will be accessible as a top menu.
        :param category_icon:
            Font-Awesome icon name for the category, optional.
        :param category_label:
            The label that will be displayed on the menu,
            if absent param name will be used
        :param baseview:
            A BaseView type class instantiated.
        :param cond:
            If a callable, :code:`cond` will be invoked when
            constructing the menu items. If it returns :code:`True`,
            then this link will be a part of the menu. Otherwise, it
            will not be included in the menu items. Defaults to
            :code:`None`, meaning the item will always be present.
        """
        self.menu.add_link(
            name=name,
            href=href,
            icon=icon,
            label=label,
            category=category,
            category_icon=category_icon,
            category_label=category_label,
            baseview=baseview,
            cond=cond,
        )
        if current_app:
            self._add_permissions_menu(name)
            if category:
                self._add_permissions_menu(category)

    def add_separator(self, category, cond=None):
        """
        Add a separator to the menu, you will sequentially create the menu.

        :param category:
            The menu category where the separator will be included.
        :param cond:
            If a callable, :code:`cond` will be invoked when
            constructing the menu items. If it returns :code:`True`,
            then this separator will be a part of the menu. Otherwise,
            it will not be included in the menu items. Defaults to
            :code:`None`, meaning the separator will always be present.
        """
        self.menu.add_separator(category, cond=cond)

    def add_view_no_menu(self, baseview, endpoint=None, static_folder=None):
        """
        Add your views without creating a menu.

        :param baseview: A BaseView type class instantiated.
        """
        baseview = self._check_and_init(baseview)
        log.debug(LOGMSG_INF_FAB_ADD_VIEW, baseview.__class__.__name__, "")

        if not self._view_exists(baseview):
            baseview.appbuilder = self
            self.baseviews.append(baseview)
            self._process_inner_views()
            self.register_blueprint(baseview, endpoint=endpoint, static_folder=static_folder)
            self._add_permission(baseview)
        else:
            log.warning(LOGMSG_WAR_FAB_VIEW_EXISTS, baseview.__class__.__name__)
        return baseview

    def add_api(self, baseview):
        """
        Add a BaseApi class or child to AppBuilder.

        :param baseview: A BaseApi type class
        """
        return self.add_view_no_menu(baseview)

    @property
    def get_url_for_index(self):
        return url_for(f"{self.indexview.endpoint}.{self.indexview.default_view}")

    def get_url_for_login_with(self, next_url: str | None = None) -> str:
        return get_auth_manager().get_url_login(next_url=next_url)

    @property
    def get_url_for_login(self):
        return get_auth_manager().get_url_login()

    def get_url_for_locale(self, lang):
        return url_for(
            f"{self.bm.locale_view.endpoint}.{self.bm.locale_view.default_view}",
            locale=lang,
        )

    def add_limits(self, baseview) -> None:
        if hasattr(baseview, "limits"):
            self.sm.add_limit_view(baseview)

    def _add_permission(self, baseview, update_perms=False):
        if self.update_perms or update_perms:
            try:
                if hasattr(self.sm, "add_permissions_view"):
                    self.sm.add_permissions_view(baseview.base_permissions, baseview.class_permission_name)
            except Exception as e:
                log.exception(e)
                log.error(LOGMSG_ERR_FAB_ADD_PERMISSION_VIEW, e)

    def add_permissions(self, update_perms=False):
        if self.update_perms or update_perms:
            for baseview in self.baseviews:
                self._add_permission(baseview, update_perms=update_perms)
            self._add_menu_permissions(update_perms=update_perms)

    def _add_permissions_menu(self, name, update_perms=False):
        if self.update_perms or update_perms:
            try:
                if hasattr(self.sm, "add_permissions_menu"):
                    self.sm.add_permissions_menu(name)
            except Exception as e:
                log.exception(e)
                log.error(LOGMSG_ERR_FAB_ADD_PERMISSION_MENU, e)

    def _add_menu_permissions(self, update_perms=False):
        if self.update_perms or update_perms:
            for category in self.menu.get_list():
                self._add_permissions_menu(category.name, update_perms=update_perms)
                for item in category.childs:
                    # don't add permission for menu separator
                    if item.name != "-":
                        self._add_permissions_menu(item.name, update_perms=update_perms)

    def register_blueprint(self, baseview, endpoint=None, static_folder=None):
        current_app.register_blueprint(
            baseview.create_blueprint(self, endpoint=endpoint, static_folder=static_folder)
        )

    def _view_exists(self, view):
        return any(baseview.__class__ == view.__class__ for baseview in self.baseviews)

    def _process_inner_views(self):
        for view in self.baseviews:
            for inner_class in view.get_uninit_inner_views():
                for v in self.baseviews:
                    if isinstance(v, inner_class) and v not in view.get_init_inner_views():
                        view.get_init_inner_views().append(v)


def init_appbuilder(app: Flask, enable_plugins: bool) -> AirflowAppBuilder:
    """Init `Flask App Builder <https://flask-appbuilder.readthedocs.io/en/latest/>`__."""
    if settings.Session is None:
        raise RuntimeError("Session not configured. Call configure_orm() first.")
    return AirflowAppBuilder(
        app=app,
        session=settings.Session(),
        base_template="airflow/main.html",
        enable_plugins=enable_plugins,
    )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os

import jinja2
import pytest

from airflow import settings
from airflow.providers.fab.www.app import create_app

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.db import parse_and_sync_to_db
from unit.fab.auth_manager.api_endpoints.api_connexion_utils import delete_user
from unit.fab.decorators import dont_initialize_flask_app_submodules
from unit.fab.utils import client_with_login


@pytest.fixture(autouse=True, scope="module")
def session():
    settings.configure_orm()
    return settings.Session


@pytest.fixture(autouse=True, scope="module")
def examples_dag_bag(session):
    dag_bag = parse_and_sync_to_db(os.devnull, include_examples=True)
    session.commit()
    return dag_bag


@pytest.fixture(scope="module")
def app(examples_dag_bag):
    @dont_initialize_flask_app_submodules(
        skip_all_except=[
            "init_api_connexion",
            "init_appbuilder",
            "init_appbuilder_links",
            "init_appbuilder_views",
            "init_flash_views",
            "init_jinja_globals",
            "init_plugins",
            "init_airflow_session_interface",
        ]
    )
    def factory():
        with conf_vars({("fab", "auth_rate_limited"): "False"}):
            return create_app(enable_plugins=False)

    app = factory()
    app.config["WTF_CSRF_ENABLED"] = False
    app.dag_bag = examples_dag_bag
    app.jinja_env.undefined = jinja2.StrictUndefined

    security_manager = app.appbuilder.sm

    test_users = [
        {
            "username": "test_admin",
            "first_name": "test_admin_first_name",
            "last_name": "test_admin_last_name",
            "email": "test_admin@fab.org",
            "role": security_manager.find_role("Admin"),
            "password": "test_admin_password",
        },
        {
            "username": "test_user",
            "first_name": "test_user_first_name",
            "last_name": "test_user_last_name",
            "email": "test_user@fab.org",
            "role": security_manager.find_role("User"),
            "password": "test_user_password",
        },
        {
            "username": "test_viewer",
            "first_name": "test_viewer_first_name",
            "last_name": "test_viewer_last_name",
            "email": "test_viewer@fab.org",
            "role": security_manager.find_role("Viewer"),
            "password": "test_viewer_password",
        },
    ]

    for user_dict in test_users:
        if not security_manager.find_user(username=user_dict["username"]):
            security_manager.add_user(**user_dict)

    yield app

    for user_dict in test_users:
        delete_user(app, user_dict["username"])


@pytest.fixture
def user_client(app):
    return client_with_login(app, username="test_user", password="test_user")


@pytest.fixture
def viewer_client(app):
    return client_with_login(app, username="test_viewer", password="test_viewer")


@pytest.fixture
def admin_client(app):
    return client_with_login(app, username="test_admin", password="test_admin")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Permissions cleanup command."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from airflow.utils import cli as cli_utils
from airflow.utils.providers_configuration_loader import providers_configuration_loaded
from airflow.utils.session import NEW_SESSION, provide_session
from airflow.utils.strings import to_boolean

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

log = logging.getLogger(__name__)


@provide_session
def cleanup_dag_permissions(dag_id: str, session: Session = NEW_SESSION) -> None:
    """
    Clean up DAG-specific permissions from Flask-AppBuilder tables.

    When a DAG is deleted, we need to clean up the corresponding permissions
    to prevent orphaned entries in the ab_view_menu table.

    This addresses issue #50905: Deleted DAGs not removed from ab_view_menu table
    and show up in permissions.

    :param dag_id: Specific DAG ID to clean up.
    :param session: Database session.
    """
    from sqlalchemy import delete, select

    from airflow.providers.fab.auth_manager.models import Permission, Resource, assoc_permission_role
    from airflow.security.permissions import RESOURCE_DAG_PREFIX, RESOURCE_DAG_RUN, RESOURCE_DETAILS_MAP

    # Clean up specific DAG permissions
    dag_resources = session.scalars(
        select(Resource).filter(
            Resource.name.in_(
                [
                    f"{RESOURCE_DAG_PREFIX}{dag_id}",  # DAG:dag_id
                    f"{RESOURCE_DETAILS_MAP[RESOURCE_DAG_RUN]['prefix']}{dag_id}",  # DAG_RUN:dag_id
                ]
            )
        )
    ).all()
    log.info("Cleaning up DAG-specific permissions for dag_id: %s", dag_id)

    if not dag_resources:
        return

    dag_resource_ids = [resource.id for resource in dag_resources]

    # Find all permissions associated with these resources
    dag_permissions = session.scalars(
        select(Permission).filter(Permission.resource_id.in_(dag_resource_ids))
    ).all()

    if not dag_permissions:
        # Delete resources even if no permissions exist
        session.execute(delete(Resource).where(Resource.id.in_(dag_resource_ids)))
        return

    dag_permission_ids = [permission.id for permission in dag_permissions]

    # Delete permission-role associations first (foreign key constraint)
    session.execute(
        delete(assoc_permission_role).where(
            assoc_permission_role.c.permission_view_id.in_(dag_permission_ids)
        )
    )

    # Delete permissions
    session.execute(delete(Permission).where(Permission.resource_id.in_(dag_resource_ids)))

    # Delete resources (ab_view_menu entries)
    session.execute(delete(Resource).where(Resource.id.in_(dag_resource_ids)))

    log.info("Cleaned up %d DAG-specific permissions", len(dag_permissions))


@cli_utils.action_cli
@providers_configuration_loaded
def permissions_cleanup(args):
    """Clean up DAG permissions in Flask-AppBuilder tables."""
    from sqlalchemy import select

    from airflow.models import DagModel
    from airflow.providers.fab.auth_manager.cli_commands.utils import get_application_builder
    from airflow.providers.fab.auth_manager.models import Resource
    from airflow.security.permissions import (
        RESOURCE_DAG_PREFIX,
        RESOURCE_DAG_RUN,
        RESOURCE_DETAILS_MAP,
    )
    from airflow.utils.session import create_session

    with get_application_builder() as _:
        with create_session() as session:
            # Get all existing DAG IDs from DagModel
            existing_dag_ids = {dag.dag_id for dag in session.scalars(select(DagModel)).all()}

            # Get all DAG-related resources from FAB tables
            dag_resources = session.scalars(
                select(Resource).filter(
                    Resource.name.like(f"{RESOURCE_DAG_PREFIX}%")
                    | Resource.name.like(f"{RESOURCE_DETAILS_MAP[RESOURCE_DAG_RUN]['prefix']}%")
                )
            ).all()

            orphaned_resources = []
            orphaned_dag_ids = set()

            for resource in dag_resources:
                # Extract DAG ID from resource name
                dag_id = None
                if resource.name.startswith(RESOURCE_DAG_PREFIX):
                    dag_id = resource.name[len(RESOURCE_DAG_PREFIX) :]
                elif resource.name.startswith(RESOURCE_DETAILS_MAP[RESOURCE_DAG_RUN]["prefix"]):
                    dag_id = resource.name[len(RESOURCE_DETAILS_MAP[RESOURCE_DAG_RUN]["prefix"]) :]

                # Check if this DAG ID still exists
                if dag_id and dag_id not in existing_dag_ids:
                    orphaned_resources.append(resource)
                    orphaned_dag_ids.add(dag_id)

            # Filter by specific DAG ID if provided
            if args.dag_id:
                if args.dag_id in orphaned_dag_ids:
                    orphaned_dag_ids = {args.dag_id}
                    print(f"Filtering to clean up permissions for DAG: {args.dag_id}")
                else:
                    print(
                        f"DAG '{args.dag_id}' not found in orphaned permissions or still exists in database."
                    )
                    return

            if not orphaned_dag_ids:
                if args.dag_id:
                    print(f"No orphaned permissions found for DAG: {args.dag_id}")
                else:
                    print("No orphaned DAG permissions found.")
                return

            print(f"Found orphaned permissions for {len(orphaned_dag_ids)} deleted DAG(s):")
            for dag_id in sorted(orphaned_dag_ids):
                print(f"  - {dag_id}")

            if args.dry_run:
                print("\nDry run mode: No changes will be made.")
                print(f"Would clean up permissions for {len(orphaned_dag_ids)} orphaned DAG(s).")
                return

            # Perform cleanup if not in dry run mode
            if not args.yes:
                action = (
                    f"clean up permissions for {len(orphaned_dag_ids)} DAG(s)"
                    if not args.dag_id
                    else f"clean up permissions for DAG '{args.dag_id}'"
                )
                confirm = input(f"\nDo you want to {action}? [y/N]: ")
                if not to_boolean(confirm):
                    print("Cleanup cancelled.")
                    return

            # Perform the actual cleanup
            cleanup_count = 0
            for dag_id in orphaned_dag_ids:
                try:
                    cleanup_dag_permissions(dag_id, session)
                    cleanup_count += 1
                    if args.verbose:
                        print(f"Cleaned up permissions for DAG: {dag_id}")
                except Exception as e:
                    print(f"Failed to clean up permissions for DAG {dag_id}: {e}")

            print(f"\nSuccessfully cleaned up permissions for {cleanup_count} DAG(s).")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest.mock import ANY, MagicMock, patch

import pytest
from flask_appbuilder.const import AUTH_DB, AUTH_LDAP
from starlette.exceptions import HTTPException

from airflow.providers.fab.auth_manager.api_fastapi.services.login import FABAuthManagerLogin


@pytest.fixture
def auth_manager():
    return MagicMock()


@pytest.fixture
def security_manager():
    return MagicMock()


@pytest.fixture
def user():
    user = MagicMock()
    user.password = "dummy"
    return user


@patch("airflow.providers.fab.auth_manager.api_fastapi.services.login.get_auth_manager")
class TestLogin:
    def setup_method(
        self,
    ):
        self.login_body = MagicMock()
        self.login_body.username = "username"
        self.login_body.password = "password"
        self.dummy_token = "DUMMY_TOKEN"

    @pytest.mark.parametrize(
        ("auth_type", "method"),
        [
            [AUTH_DB, "auth_user_db"],
            [AUTH_LDAP, "auth_user_ldap"],
        ],
    )
    def test_create_token(self, get_auth_manager, auth_type, method, auth_manager, security_manager, user):
        security_manager.auth_type = auth_type
        getattr(security_manager, method).return_value = user

        auth_manager.security_manager = security_manager
        auth_manager.generate_jwt.return_value = self.dummy_token

        get_auth_manager.return_value = auth_manager

        result = FABAuthManagerLogin.create_token(
            body=self.login_body,
        )
        assert result.access_token == self.dummy_token
        getattr(security_manager, method).assert_called_once_with(
            self.login_body.username, self.login_body.password, rotate_session_id=False
        )
        auth_manager.generate_jwt.assert_called_once_with(user=user, expiration_time_in_seconds=ANY)

    @pytest.mark.parametrize(
        ("auth_type", "methods"),
        [
            [AUTH_DB, ["auth_user_db"]],
            [AUTH_LDAP, ["auth_user_ldap", "auth_user_db"]],
        ],
    )
    def test_create_token_no_user(
        self, get_auth_manager, auth_type, methods, auth_manager, security_manager, user
    ):
        security_manager.auth_type = auth_type
        for method in methods:
            getattr(security_manager, method).return_value = None

        auth_manager.security_manager = security_manager
        get_auth_manager.return_value = auth_manager

        with pytest.raises(HTTPException) as ex:
            FABAuthManagerLogin.create_token(
                body=self.login_body,
            )
        assert ex.value.status_code == 401



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import Mock, patch

import pytest
from flask import Response
from flask_appbuilder.const import AUTH_LDAP

from airflow.providers.fab.auth_manager.api.auth.backend.basic_auth import requires_authentication
from airflow.providers.fab.www import app as application


@pytest.fixture
def app():
    return application.create_app(enable_plugins=False)


@pytest.fixture
def mock_sm():
    return Mock()


@pytest.fixture
def mock_auth_manager(mock_sm):
    auth_manager = Mock()
    auth_manager.security_manager = mock_sm
    return auth_manager


@pytest.fixture
def mock_authorization():
    authorization = Mock()
    authorization.username = "username"
    authorization.password = "password"
    return authorization


mock_call = Mock()


@requires_authentication
def function_decorated():
    mock_call()


@pytest.mark.db_test
class TestBasicAuth:
    def setup_method(self) -> None:
        mock_call.reset_mock()

    def test_requires_authentication_with_no_header(self, app):
        with app.test_request_context() as mock_context:
            mock_context.request.authorization = None
            result = function_decorated()

        assert type(result) is Response
        assert result.status_code == 401

    @patch("airflow.providers.fab.auth_manager.api.auth.backend.basic_auth.get_auth_manager")
    @patch("airflow.providers.fab.auth_manager.api.auth.backend.basic_auth.login_user")
    def test_requires_authentication_with_ldap(
        self, mock_login_user, mock_get_auth_manager, app, mock_sm, mock_auth_manager, mock_authorization
    ):
        mock_sm.auth_type = AUTH_LDAP
        mock_get_auth_manager.return_value = mock_auth_manager
        user = Mock()
        mock_sm.auth_user_ldap.return_value = user

        with app.test_request_context() as mock_context:
            mock_context.request.authorization = mock_authorization
            function_decorated()

        mock_sm.auth_user_ldap.assert_called_once_with(
            mock_authorization.username, mock_authorization.password
        )
        mock_login_user.assert_called_once_with(user, remember=False)
        mock_call.assert_called_once()

    @patch("airflow.providers.fab.auth_manager.api.auth.backend.basic_auth.get_auth_manager")
    @patch("airflow.providers.fab.auth_manager.api.auth.backend.basic_auth.login_user")
    def test_requires_authentication_with_db(
        self, mock_login_user, mock_get_auth_manager, app, mock_sm, mock_auth_manager, mock_authorization
    ):
        mock_get_auth_manager.return_value = mock_auth_manager
        user = Mock()
        mock_sm.auth_user_db.return_value = user

        with app.test_request_context() as mock_context:
            mock_context.request.authorization = mock_authorization
            function_decorated()

        mock_sm.auth_user_db.assert_called_once_with(mock_authorization.username, mock_authorization.password)
        mock_login_user.assert_called_once_with(user, remember=False)
        mock_call.assert_called_once()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os

import pytest

import airflow
from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException
from airflow.providers.fab.auth_manager.cli_commands.utils import get_application_builder
from airflow.providers.fab.www.extensions.init_appbuilder import AirflowAppBuilder
from airflow.providers.fab.www.session import AirflowDatabaseSessionInterface

from tests_common.test_utils.config import conf_vars


@pytest.fixture
def flask_app():
    """Fixture to set up the Flask app with the necessary configuration."""
    # Get the webserver config file path
    webserver_config = conf.get_mandatory_value("fab", "config_file")
    with get_application_builder() as appbuilder:
        flask_app = appbuilder.app

        # Load webserver configuration
        flask_app.config.from_pyfile(webserver_config, silent=True)

        yield flask_app


@pytest.mark.db_test
class TestCliUtils:
    def test_get_application_builder(self):
        """Test that get_application_builder returns an AirflowAppBuilder instance."""
        with get_application_builder() as appbuilder:
            assert isinstance(appbuilder, AirflowAppBuilder)

    def test_sqlalchemy_uri_configured(self, flask_app):
        """Test that the SQLALCHEMY_DATABASE_URI is correctly set in the Flask app."""
        sqlalchemy_uri = conf.get("database", "SQL_ALCHEMY_CONN")

        # Assert that the SQLAlchemy URI is correctly set
        assert sqlalchemy_uri == flask_app.config["SQLALCHEMY_DATABASE_URI"]

    def test_relative_path_sqlite_raises_exception(self):
        """Test that a relative SQLite path raises an AirflowConfigException."""
        # Directly simulate the configuration for relative SQLite path
        with conf_vars({("database", "SQL_ALCHEMY_CONN"): "sqlite://relative/path"}):
            with pytest.raises(AirflowConfigException, match="Cannot use relative path"):
                with get_application_builder():
                    pass

    def test_static_folder_exists(self, flask_app):
        """Test that the static folder is correctly configured in the Flask app."""
        static_folder = os.path.join(os.path.dirname(airflow.__file__), "www", "static")
        assert flask_app.static_folder == static_folder

    def test_database_auth_backend_in_session(self, flask_app):
        """Test that the database is used for session management when AUTH_BACKEND is set to 'database'."""

        with get_application_builder() as appbuilder:
            flask_app = appbuilder.app
            # Ensure that the correct session interface is set (for 'database' auth backend)
            assert isinstance(flask_app.session_interface, AirflowDatabaseSessionInterface)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from contextlib import contextmanager, suppress
from itertools import chain
from typing import TYPE_CHECKING
from unittest import mock
from unittest.mock import MagicMock, Mock

import pytest
from flask import g

from airflow.api_fastapi.app import AUTH_MANAGER_FASTAPI_APP_PREFIX, get_auth_manager
from airflow.api_fastapi.common.types import MenuItem
from airflow.exceptions import AirflowConfigException
from airflow.providers.fab.www.app import create_app
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.utils.db import resetdb

from tests_common.test_utils.config import conf_vars
from unit.fab.auth_manager.api_endpoints.api_connexion_utils import create_user, delete_user

with suppress(ImportError):
    from airflow.api_fastapi.auth.managers.models.resource_details import (
        AccessView,
        DagAccessEntity,
        DagDetails,
    )

from airflow.providers.common.compat.security.permissions import (
    RESOURCE_ASSET,
    RESOURCE_ASSET_ALIAS,
    RESOURCE_BACKFILL,
)
from airflow.providers.fab.auth_manager.fab_auth_manager import FabAuthManager
from airflow.providers.fab.auth_manager.security_manager.override import FabAirflowSecurityManagerOverride
from airflow.providers.fab.www.security.permissions import (
    ACTION_CAN_ACCESS_MENU,
    ACTION_CAN_CREATE,
    ACTION_CAN_DELETE,
    ACTION_CAN_EDIT,
    ACTION_CAN_READ,
    RESOURCE_AUDIT_LOG,
    RESOURCE_CONFIG,
    RESOURCE_CONNECTION,
    RESOURCE_DAG,
    RESOURCE_DAG_RUN,
    RESOURCE_DOCS,
    RESOURCE_JOB,
    RESOURCE_PLUGIN,
    RESOURCE_PROVIDER,
    RESOURCE_TASK_INSTANCE,
    RESOURCE_TRIGGER,
    RESOURCE_VARIABLE,
    RESOURCE_WEBSITE,
)

from tests_common.test_utils.dag import sync_dag_to_db
from tests_common.test_utils.version_compat import AIRFLOW_V_3_1_PLUS

if AIRFLOW_V_3_1_PLUS:
    from airflow.providers.fab.www.security.permissions import RESOURCE_HITL_DETAIL

    HITL_ENDPOINT_TESTS = [
        # With global permissions on Dags, but no permission on HITL Detail
        (
            "GET",
            DagAccessEntity.HITL_DETAIL,
            None,
            [(ACTION_CAN_READ, RESOURCE_DAG)],
            False,
        ),
        # With global permissions on Dags, but no permission on HITL Detail
        (
            "PUT",
            DagAccessEntity.HITL_DETAIL,
            None,
            [(ACTION_CAN_READ, RESOURCE_DAG)],
            False,
        ),
        # With global permissions on Dags, with read permission on HITL Detail
        (
            "GET",
            DagAccessEntity.HITL_DETAIL,
            None,
            [(ACTION_CAN_READ, RESOURCE_DAG), (ACTION_CAN_READ, RESOURCE_HITL_DETAIL)],
            True,
        ),
        # With global permissions on Dags, with read permission on HITL Detail, but wrong method
        (
            "PUT",
            DagAccessEntity.HITL_DETAIL,
            None,
            [(ACTION_CAN_READ, RESOURCE_DAG), (ACTION_CAN_READ, RESOURCE_HITL_DETAIL)],
            False,
        ),
        # With global permissions on Dags, with write permission on HITL Detail, but wrong method
        (
            "GET",
            DagAccessEntity.HITL_DETAIL,
            None,
            [(ACTION_CAN_READ, RESOURCE_DAG), (ACTION_CAN_EDIT, RESOURCE_HITL_DETAIL)],
            False,
        ),
        # With global permissions on Dags, with edit permission on HITL Detail
        (
            "PUT",
            DagAccessEntity.HITL_DETAIL,
            None,
            [(ACTION_CAN_READ, RESOURCE_DAG), (ACTION_CAN_EDIT, RESOURCE_HITL_DETAIL)],
            True,
        ),
    ]

if TYPE_CHECKING:
    from airflow.api_fastapi.auth.managers.base_auth_manager import ResourceMethod


IS_AUTHORIZED_METHODS_SIMPLE = {
    "is_authorized_configuration": RESOURCE_CONFIG,
    "is_authorized_connection": RESOURCE_CONNECTION,
    "is_authorized_asset": RESOURCE_ASSET,
    "is_authorized_asset_alias": RESOURCE_ASSET_ALIAS,
    "is_authorized_backfill": RESOURCE_BACKFILL,
    "is_authorized_variable": RESOURCE_VARIABLE,
}


@contextmanager
def user_set(app, user):
    g.user = user
    yield
    g.user = None


@pytest.fixture
def auth_manager():
    return FabAuthManager(None)


@pytest.fixture
def flask_app():
    with conf_vars(
        {
            (
                "core",
                "auth_manager",
            ): "airflow.providers.fab.auth_manager.fab_auth_manager.FabAuthManager",
        }
    ):
        app = create_app(enable_plugins=False)
        with app.app_context():
            yield app


@pytest.fixture
def auth_manager_with_appbuilder(flask_app):
    return get_auth_manager()


@pytest.mark.db_test
class TestFabAuthManager:
    @mock.patch("flask_login.utils._get_user")
    def test_get_user(self, mock_current_user, minimal_app_for_auth_api, auth_manager):
        user = Mock()
        user.is_anonymous.return_value = True
        mock_current_user.return_value = user
        with minimal_app_for_auth_api.app_context():
            assert auth_manager.get_user() == user

    @mock.patch("flask_login.utils._get_user")
    def test_get_user_from_flask_g(self, mock_current_user, minimal_app_for_auth_api, auth_manager):
        session_user = Mock()
        session_user.is_anonymous = True
        mock_current_user.return_value = session_user

        flask_g_user = Mock()
        flask_g_user.is_anonymous = False
        with minimal_app_for_auth_api.app_context():
            with user_set(minimal_app_for_auth_api, flask_g_user):
                assert auth_manager.get_user() == flask_g_user

    def test_deserialize_user(self, flask_app, auth_manager_with_appbuilder):
        user = create_user(flask_app, "test")
        result = auth_manager_with_appbuilder.deserialize_user({"sub": str(user.id)})
        assert user.get_id() == result.get_id()

    def test_serialize_user(self, flask_app, auth_manager_with_appbuilder):
        user = create_user(flask_app, "test")
        result = auth_manager_with_appbuilder.serialize_user(user)
        assert result == {"sub": str(user.id)}

    @mock.patch.object(FabAuthManager, "get_user")
    def test_is_logged_in(self, mock_get_user, auth_manager_with_appbuilder):
        user = Mock()
        user.is_anonymous.return_value = True
        mock_get_user.return_value = user

        assert auth_manager_with_appbuilder.is_logged_in() is False

    @mock.patch.object(FabAuthManager, "get_user")
    def test_is_logged_in_with_inactive_user(self, mock_get_user, auth_manager_with_appbuilder):
        user = Mock()
        user.is_anonymous.return_value = False
        user.is_active.return_value = True
        mock_get_user.return_value = user

        assert auth_manager_with_appbuilder.is_logged_in() is False

    @pytest.mark.parametrize(
        ("api_name", "method", "user_permissions", "expected_result"),
        chain(
            *[
                (
                    # With permission
                    (
                        api_name,
                        "POST",
                        [(ACTION_CAN_CREATE, resource_type)],
                        True,
                    ),
                    # With permission
                    (
                        api_name,
                        "GET",
                        [(ACTION_CAN_READ, resource_type)],
                        True,
                    ),
                    # With permission (with several user permissions)
                    (
                        api_name,
                        "DELETE",
                        [(ACTION_CAN_DELETE, resource_type), (ACTION_CAN_CREATE, "resource_test")],
                        True,
                    ),
                    # With permission
                    (
                        api_name,
                        "MENU",
                        [(ACTION_CAN_ACCESS_MENU, resource_type)],
                        True,
                    ),
                    # Without permission
                    (
                        api_name,
                        "POST",
                        [(ACTION_CAN_READ, resource_type), (ACTION_CAN_CREATE, "resource_test")],
                        False,
                    ),
                )
                for api_name, resource_type in IS_AUTHORIZED_METHODS_SIMPLE.items()
            ]
        ),
    )
    def test_is_authorized(self, api_name, method, user_permissions, expected_result, auth_manager):
        user = Mock()
        user.perms = user_permissions
        result = getattr(auth_manager, api_name)(
            method=method,
            user=user,
        )
        assert result == expected_result

    @pytest.mark.parametrize(
        ("method", "dag_access_entity", "dag_details", "user_permissions", "expected_result"),
        [
            # Scenario 1 #
            # With global permissions on Dags
            (
                "GET",
                None,
                None,
                [(ACTION_CAN_READ, RESOURCE_DAG)],
                True,
            ),
            # Without permission on DAGs
            (
                "GET",
                None,
                None,
                [(ACTION_CAN_READ, "resource_test")],
                False,
            ),
            # With specific DAG permissions but no specific DAG requested
            (
                "GET",
                None,
                None,
                [(ACTION_CAN_READ, "DAG:test_dag_id")],
                True,
            ),
            # With multiple specific DAG permissions, no specific DAG requested
            (
                "GET",
                None,
                None,
                [(ACTION_CAN_READ, "DAG:test_dag_id"), (ACTION_CAN_READ, "DAG:test_dag_id2")],
                True,
            ),
            # With specific DAG permissions and wrong method
            (
                "POST",
                None,
                None,
                [(ACTION_CAN_READ, "DAG:test_dag_id")],
                False,
            ),
            # With correct POST permissions
            (
                "POST",
                None,
                None,
                [(ACTION_CAN_CREATE, RESOURCE_DAG)],
                True,
            ),
            # Mixed permissions - some DAG, some non-DAG
            (
                "GET",
                None,
                None,
                [(ACTION_CAN_READ, "DAG:test_dag_id"), (ACTION_CAN_READ, "resource_test")],
                True,
            ),
            # DAG sub-entity with specific DAG permissions but no specific DAG requested
            (
                "GET",
                DagAccessEntity.RUN,
                None,
                [(ACTION_CAN_READ, "DAG:test_dag_id"), (ACTION_CAN_READ, RESOURCE_DAG_RUN)],
                True,
            ),
            # DAG sub-entity access with no DAG permissions, no specific DAG requested
            (
                "GET",
                DagAccessEntity.RUN,
                None,
                [(ACTION_CAN_READ, RESOURCE_DAG_RUN)],
                False,
            ),
            # DAG sub-entity with specific DAG permissions but missing sub-entity permission
            (
                "GET",
                DagAccessEntity.TASK_INSTANCE,
                None,
                [(ACTION_CAN_READ, "DAG:test_dag_id")],
                False,
            ),
            # Multiple DAG access entities with proper permissions
            (
                "DELETE",
                DagAccessEntity.TASK,
                None,
                [(ACTION_CAN_EDIT, "DAG:test_dag_id"), (ACTION_CAN_DELETE, RESOURCE_TASK_INSTANCE)],
                True,
            ),
            # User with specific DAG permissions but wrong method for sub-entity
            (
                "POST",
                DagAccessEntity.RUN,
                None,
                [(ACTION_CAN_READ, "DAG:test_dag_id"), (ACTION_CAN_READ, RESOURCE_DAG_RUN)],
                False,
            ),
            # Scenario 2 #
            # On specific DAG with global permissions on Dags
            (
                "GET",
                None,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_READ, RESOURCE_DAG)],
                True,
            ),
            # With permission on a specific DAG
            (
                "GET",
                None,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_READ, "DAG:test_dag_id"), (ACTION_CAN_READ, "DAG:test_dag_id2")],
                True,
            ),
            # Without permission on a specific DAG (wrong method)
            (
                "POST",
                None,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_READ, "DAG:test_dag_id")],
                False,
            ),
            # Without permission on a specific DAG
            (
                "GET",
                None,
                DagDetails(id="test_dag_id2"),
                [(ACTION_CAN_READ, "DAG:test_dag_id")],
                False,
            ),
            # Scenario 3 #
            # With global permissions on DAGs
            (
                "GET",
                DagAccessEntity.RUN,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_READ, RESOURCE_DAG), (ACTION_CAN_READ, RESOURCE_DAG_RUN)],
                True,
            ),
            # Without read permissions on a specific DAG
            (
                "GET",
                DagAccessEntity.TASK_INSTANCE,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_READ, RESOURCE_TASK_INSTANCE)],
                False,
            ),
            # With read permissions on a specific DAG but not on the DAG run
            (
                "GET",
                DagAccessEntity.TASK_INSTANCE,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_READ, "DAG:test_dag_id"), (ACTION_CAN_READ, RESOURCE_TASK_INSTANCE)],
                False,
            ),
            # With read permissions on a specific DAG and on the DAG run
            (
                "GET",
                DagAccessEntity.TASK_INSTANCE,
                DagDetails(id="test_dag_id"),
                [
                    (ACTION_CAN_READ, "DAG:test_dag_id"),
                    (ACTION_CAN_READ, RESOURCE_TASK_INSTANCE),
                    (ACTION_CAN_READ, RESOURCE_DAG_RUN),
                ],
                True,
            ),
            # With edit permissions on a specific DAG and delete on the DAG access entity
            (
                "DELETE",
                DagAccessEntity.TASK,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_EDIT, "DAG:test_dag_id"), (ACTION_CAN_DELETE, RESOURCE_TASK_INSTANCE)],
                True,
            ),
            # With edit permissions on a specific DAG and create on the DAG access entity
            (
                "POST",
                DagAccessEntity.RUN,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_EDIT, "DAG:test_dag_id"), (ACTION_CAN_CREATE, RESOURCE_DAG_RUN)],
                True,
            ),
            # Without permissions to edit the DAG
            (
                "POST",
                DagAccessEntity.RUN,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_CREATE, RESOURCE_DAG_RUN)],
                False,
            ),
            # Without read permissions on a specific DAG
            (
                "GET",
                DagAccessEntity.TASK_LOGS,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_READ, RESOURCE_TASK_INSTANCE)],
                False,
            ),
            # Use deprecated prefix "DAG Run" to assign permissions specifically on dag runs
            (
                "GET",
                DagAccessEntity.RUN,
                DagDetails(id="test_dag_id"),
                [(ACTION_CAN_READ, "DAG:test_dag_id"), (ACTION_CAN_READ, "DAG Run:test_dag_id")],
                True,
            ),
        ],
    )
    def test_is_authorized_dag(
        self,
        method,
        dag_access_entity,
        dag_details,
        user_permissions,
        expected_result,
        auth_manager_with_appbuilder,
    ):
        user = Mock()
        user.perms = user_permissions
        user.id = 1
        result = auth_manager_with_appbuilder.is_authorized_dag(
            method=method, access_entity=dag_access_entity, details=dag_details, user=user
        )
        assert result == expected_result

    @pytest.mark.skipif(
        AIRFLOW_V_3_1_PLUS is not True, reason="HITL test will be skipped if Airflow version < 3.1.0"
    )
    @pytest.mark.parametrize(
        ("method", "dag_access_entity", "dag_details", "user_permissions", "expected_result"),
        HITL_ENDPOINT_TESTS if AIRFLOW_V_3_1_PLUS else [],
    )
    @mock.patch.object(FabAuthManager, "get_authorized_dag_ids")
    def test_is_authorized_dag_hitl_detail(
        self,
        mock_get_authorized_dag_ids,
        method,
        dag_access_entity,
        dag_details,
        user_permissions,
        expected_result,
        auth_manager_with_appbuilder,
    ):
        dag_permissions = [perm[1] for perm in user_permissions if perm[1].startswith("DAG:")]
        dag_ids = {perm.replace("DAG:", "") for perm in dag_permissions}
        mock_get_authorized_dag_ids.return_value = dag_ids

        user = Mock()
        user.perms = user_permissions
        user.id = 1
        result = auth_manager_with_appbuilder.is_authorized_dag(
            method=method, access_entity=dag_access_entity, details=dag_details, user=user
        )
        assert result == expected_result

    @pytest.mark.parametrize(
        ("access_view", "user_permissions", "expected_result"),
        [
            # With permission (jobs)
            (
                AccessView.JOBS,
                [(ACTION_CAN_READ, RESOURCE_JOB)],
                True,
            ),
            # With permission (plugins)
            (
                AccessView.PLUGINS,
                [(ACTION_CAN_READ, RESOURCE_PLUGIN)],
                True,
            ),
            # With permission (providers)
            (
                AccessView.PROVIDERS,
                [(ACTION_CAN_READ, RESOURCE_PROVIDER)],
                True,
            ),
            # With permission (triggers)
            (
                AccessView.TRIGGERS,
                [(ACTION_CAN_READ, RESOURCE_TRIGGER)],
                True,
            ),
            # With permission (website)
            (
                AccessView.WEBSITE,
                [(ACTION_CAN_READ, RESOURCE_WEBSITE)],
                True,
            ),
            # Without permission
            (
                AccessView.WEBSITE,
                [(ACTION_CAN_READ, "resource_test"), (ACTION_CAN_CREATE, RESOURCE_WEBSITE)],
                False,
            ),
            # Without permission
            (
                AccessView.WEBSITE,
                [(ACTION_CAN_READ, RESOURCE_TRIGGER)],
                False,
            ),
            # Docs (positive)
            (
                AccessView.DOCS,
                [(ACTION_CAN_ACCESS_MENU, RESOURCE_DOCS)],
                True,
            ),
            # Without permission
            (
                AccessView.DOCS,
                [(ACTION_CAN_READ, RESOURCE_DOCS)],
                False,
            ),
        ],
    )
    def test_is_authorized_view(self, access_view, user_permissions, expected_result, auth_manager):
        user = Mock()
        user.perms = user_permissions
        result = auth_manager.is_authorized_view(access_view=access_view, user=user)
        assert result == expected_result

    @pytest.mark.parametrize(
        ("method", "resource_name", "user_permissions", "expected_result"),
        [
            (
                "GET",
                "custom_resource",
                [(ACTION_CAN_READ, "custom_resource")],
                True,
            ),
            (
                "GET",
                "custom_resource",
                [(ACTION_CAN_EDIT, "custom_resource")],
                False,
            ),
            (
                "GET",
                "custom_resource",
                [(ACTION_CAN_READ, "custom_resource2")],
                False,
            ),
            (
                "DUMMY",
                "custom_resource",
                [("DUMMY", "custom_resource")],
                True,
            ),
        ],
    )
    def test_is_authorized_custom_view(
        self,
        method: ResourceMethod | str,
        resource_name: str,
        user_permissions,
        expected_result,
        auth_manager,
    ):
        user = Mock()
        user.perms = user_permissions
        result = auth_manager.is_authorized_custom_view(method=method, resource_name=resource_name, user=user)
        assert result == expected_result

    @pytest.mark.parametrize(
        ("menu_items", "user_permissions", "expected_result"),
        [
            (
                [MenuItem.ASSETS, MenuItem.DAGS],
                [(ACTION_CAN_ACCESS_MENU, RESOURCE_ASSET), (ACTION_CAN_ACCESS_MENU, RESOURCE_DAG)],
                [MenuItem.ASSETS, MenuItem.DAGS],
            ),
            (
                [MenuItem.ASSETS, MenuItem.DAGS],
                [(ACTION_CAN_READ, RESOURCE_ASSET), (ACTION_CAN_READ, RESOURCE_DAG)],
                [],
            ),
            (
                [MenuItem.AUDIT_LOG, MenuItem.VARIABLES],
                [(ACTION_CAN_ACCESS_MENU, RESOURCE_AUDIT_LOG), (ACTION_CAN_READ, RESOURCE_VARIABLE)],
                [MenuItem.AUDIT_LOG],
            ),
            (
                [],
                [],
                [],
            ),
        ],
    )
    def test_filter_authorized_menu_items(
        self,
        menu_items: list[MenuItem],
        user_permissions,
        expected_result,
        auth_manager,
    ):
        user = Mock()
        user.perms = user_permissions
        result = auth_manager.filter_authorized_menu_items(menu_items, user=user)
        assert result == expected_result

    def test_get_authorized_connections(self, auth_manager):
        session = Mock()
        session.execute.return_value.scalars.return_value.all.return_value = ["conn1", "conn2"]
        result = auth_manager.get_authorized_connections(user=Mock(), method="GET", session=session)
        assert result == {"conn1", "conn2"}

    @pytest.mark.parametrize(
        ("method", "user_permissions", "expected_results"),
        [
            # Scenario 1
            # With global read permissions on Dags
            (
                "GET",
                [(ACTION_CAN_READ, RESOURCE_DAG)],
                {"test_dag1", "test_dag2", "Connections"},
            ),
            # Scenario 2
            # With global edit permissions on Dags
            (
                "PUT",
                [(ACTION_CAN_EDIT, RESOURCE_DAG)],
                {"test_dag1", "test_dag2", "Connections"},
            ),
            # Scenario 3
            # With DAG-specific permissions
            (
                "GET",
                [(ACTION_CAN_READ, "DAG:test_dag1")],
                {"test_dag1"},
            ),
            # Scenario 4
            # With no permissions
            (
                "GET",
                [],
                set(),
            ),
            # Scenario 5
            # With read permissions but edit is requested
            (
                "PUT",
                [(ACTION_CAN_READ, RESOURCE_DAG)],
                set(),
            ),
            # Scenario 7
            # With read permissions but edit is requested
            (
                "PUT",
                [(ACTION_CAN_READ, "DAG:test_dag1")],
                set(),
            ),
            # Scenario 8
            # With DAG-specific permissions
            (
                "PUT",
                [(ACTION_CAN_EDIT, "DAG:test_dag1"), (ACTION_CAN_EDIT, "DAG:test_dag2")],
                {"test_dag1", "test_dag2"},
            ),
            # Scenario 9
            # With non-DAG related permissions
            (
                "GET",
                [(ACTION_CAN_READ, "DAG:test_dag1"), (ACTION_CAN_READ, RESOURCE_CONNECTION)],
                {"test_dag1"},
            ),
        ],
    )
    def test_get_authorized_dag_ids(
        self, method, user_permissions, expected_results, auth_manager_with_appbuilder, flask_app, dag_maker
    ):
        with dag_maker("test_dag1"):
            EmptyOperator(task_id="task1")
        if AIRFLOW_V_3_1_PLUS:
            sync_dag_to_db(dag_maker.dag)
        with dag_maker("test_dag2"):
            EmptyOperator(task_id="task2")
        if AIRFLOW_V_3_1_PLUS:
            sync_dag_to_db(dag_maker.dag)
        with dag_maker("Connections"):
            EmptyOperator(task_id="task3")
        if AIRFLOW_V_3_1_PLUS:
            sync_dag_to_db(dag_maker.dag)
        dag_maker.session.commit()
        dag_maker.session.close()

        user = create_user(
            flask_app,
            username="username",
            role_name="test",
            permissions=user_permissions,
        )

        auth_manager_with_appbuilder.security_manager.sync_perm_for_dag("test_dag1")
        auth_manager_with_appbuilder.security_manager.sync_perm_for_dag("test_dag2")

        results = auth_manager_with_appbuilder.get_authorized_dag_ids(user=user, method=method)
        assert results == expected_results

        delete_user(flask_app, "username")

    def test_get_authorized_pools(self, auth_manager):
        session = Mock()
        session.execute.return_value.scalars.return_value.all.return_value = ["pool1", "pool2"]
        result = auth_manager.get_authorized_pools(user=Mock(), method="GET", session=session)
        assert result == {"pool1", "pool2"}

    def test_get_authorized_variables(self, auth_manager):
        session = Mock()
        session.execute.return_value.scalars.return_value.all.return_value = ["var1", "var2"]
        result = auth_manager.get_authorized_variables(user=Mock(), method="GET", session=session)
        assert result == {"var1", "var2"}

    def test_security_manager_return_fab_security_manager_override(self, auth_manager_with_appbuilder):
        assert isinstance(auth_manager_with_appbuilder.security_manager, FabAirflowSecurityManagerOverride)

    def test_security_manager_return_custom_provided(self, flask_app, auth_manager_with_appbuilder):
        class TestSecurityManager(FabAirflowSecurityManagerOverride):
            pass

        flask_app.config["SECURITY_MANAGER_CLASS"] = TestSecurityManager
        # Invalidate the cache
        del auth_manager_with_appbuilder.__dict__["security_manager"]
        assert isinstance(auth_manager_with_appbuilder.security_manager, TestSecurityManager)

    def test_security_manager_wrong_inheritance_raise_exception(
        self, flask_app, auth_manager_with_appbuilder
    ):
        class TestSecurityManager:
            pass

        flask_app.config["SECURITY_MANAGER_CLASS"] = TestSecurityManager
        # Invalidate the cache
        del auth_manager_with_appbuilder.__dict__["security_manager"]
        with pytest.raises(
            AirflowConfigException,
            match="Your CUSTOM_SECURITY_MANAGER must extend FabAirflowSecurityManagerOverride.",
        ):
            auth_manager_with_appbuilder.security_manager

    def test_get_url_login(self, auth_manager):
        result = auth_manager.get_url_login()
        assert result == f"{AUTH_MANAGER_FASTAPI_APP_PREFIX}/login/"

    def test_get_url_logout(self, auth_manager):
        result = auth_manager.get_url_logout()
        assert result == f"{AUTH_MANAGER_FASTAPI_APP_PREFIX}/logout"

    @mock.patch.object(FabAuthManager, "_is_authorized", return_value=True)
    def test_get_extra_menu_items(self, _, auth_manager_with_appbuilder, flask_app):
        result = auth_manager_with_appbuilder.get_extra_menu_items(user=Mock())
        assert len(result) == 5
        assert all(item.href.startswith(AUTH_MANAGER_FASTAPI_APP_PREFIX) for item in result)

    def test_get_db_manager(self, auth_manager):
        result = auth_manager.get_db_manager()
        assert result == "airflow.providers.fab.auth_manager.models.db.FABDBManager"


@pytest.mark.db_test
@pytest.mark.parametrize("skip_init", [False, True])
@conf_vars(
    {("database", "external_db_managers"): "airflow.providers.fab.auth_manager.models.db.FABDBManager"}
)
@mock.patch("airflow.providers.fab.auth_manager.models.db.FABDBManager")
@mock.patch("airflow.utils.db.create_global_lock", new=MagicMock)
@mock.patch("airflow.utils.db.drop_airflow_models")
@mock.patch("airflow.utils.db.drop_airflow_moved_tables")
@mock.patch("airflow.utils.db.initdb")
@mock.patch("airflow.settings.engine.connect")
def test_resetdb(
    mock_connect,
    mock_init,
    mock_drop_moved,
    mock_drop_airflow,
    mock_fabdb_manager,
    skip_init,
):
    session_mock = MagicMock()
    resetdb(session_mock, skip_init=skip_init)
    mock_drop_airflow.assert_called_once_with(mock_connect.return_value)
    mock_drop_moved.assert_called_once_with(mock_connect.return_value)
    if skip_init:
        mock_init.assert_not_called()
    else:
        mock_init.assert_called_once_with(session=session_mock)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import argparse
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin

import packaging.version
from connexion import FlaskApi
from fastapi import FastAPI
from flask import Blueprint, current_app, g
from sqlalchemy import select
from sqlalchemy.orm import Session, joinedload
from starlette.middleware.wsgi import WSGIMiddleware

from airflow import __version__ as airflow_version
from airflow.api_fastapi.app import AUTH_MANAGER_FASTAPI_APP_PREFIX
from airflow.api_fastapi.auth.managers.base_auth_manager import BaseAuthManager

try:
    from airflow.api_fastapi.auth.managers.base_auth_manager import ExtendedResourceMethod
except ImportError:
    from airflow.api_fastapi.auth.managers.base_auth_manager import ResourceMethod as ExtendedResourceMethod

from airflow.api_fastapi.auth.managers.models.resource_details import (
    AccessView,
    BackfillDetails,
    ConfigurationDetails,
    ConnectionDetails,
    DagAccessEntity,
    DagDetails,
    PoolDetails,
    VariableDetails,
)
from airflow.api_fastapi.common.types import ExtraMenuItem, MenuItem
from airflow.cli.cli_config import (
    DefaultHelpParser,
    GroupCommand,
)
from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException, AirflowException
from airflow.models import Connection, DagModel, Pool, Variable
from airflow.providers.fab.auth_manager.cli_commands.definition import (
    DB_COMMANDS,
    PERMISSIONS_CLEANUP_COMMAND,
    ROLES_COMMANDS,
    SYNC_PERM_COMMAND,
    USERS_COMMANDS,
)
from airflow.providers.fab.auth_manager.models import Permission, Role, User
from airflow.providers.fab.auth_manager.models.anonymous_user import AnonymousUser
from airflow.providers.fab.version_compat import AIRFLOW_V_3_1_PLUS
from airflow.providers.fab.www.app import create_app
from airflow.providers.fab.www.constants import SWAGGER_BUNDLE, SWAGGER_ENABLED
from airflow.providers.fab.www.extensions.init_views import (
    _CustomErrorRequestBodyValidator,
    _LazyResolver,
)
from airflow.providers.fab.www.security import permissions
from airflow.providers.fab.www.security.permissions import (
    ACTION_CAN_READ,
    RESOURCE_AUDIT_LOG,
    RESOURCE_CLUSTER_ACTIVITY,
    RESOURCE_CONFIG,
    RESOURCE_CONNECTION,
    RESOURCE_DAG,
    RESOURCE_DAG_CODE,
    RESOURCE_DAG_DEPENDENCIES,
    RESOURCE_DAG_PREFIX,
    RESOURCE_DAG_RUN,
    RESOURCE_DAG_VERSION,
    RESOURCE_DAG_WARNING,
    RESOURCE_DOCS,
    RESOURCE_IMPORT_ERROR,
    RESOURCE_JOB,
    RESOURCE_PLUGIN,
    RESOURCE_POOL,
    RESOURCE_PROVIDER,
    RESOURCE_TASK_INSTANCE,
    RESOURCE_TASK_LOG,
    RESOURCE_TRIGGER,
    RESOURCE_VARIABLE,
    RESOURCE_WEBSITE,
    RESOURCE_XCOM,
)
from airflow.providers.fab.www.utils import (
    get_fab_action_from_method_map,
    get_method_from_fab_action_map,
)
from airflow.security.permissions import RESOURCE_BACKFILL
from airflow.utils.session import NEW_SESSION, create_session, provide_session
from airflow.utils.yaml import safe_load

if TYPE_CHECKING:
    from airflow.api_fastapi.auth.managers.base_auth_manager import ResourceMethod
    from airflow.cli.cli_config import (
        CLICommand,
    )
    from airflow.providers.common.compat.assets import AssetAliasDetails, AssetDetails
    from airflow.providers.fab.auth_manager.security_manager.override import (
        FabAirflowSecurityManagerOverride,
    )
    from airflow.providers.fab.www.extensions.init_appbuilder import AirflowAppBuilder
    from airflow.providers.fab.www.security.permissions import (
        RESOURCE_ASSET,
        RESOURCE_ASSET_ALIAS,
    )
else:
    from airflow.providers.common.compat.security.permissions import (
        RESOURCE_ASSET,
        RESOURCE_ASSET_ALIAS,
    )


_MAP_DAG_ACCESS_ENTITY_TO_FAB_RESOURCE_TYPE: dict[DagAccessEntity, tuple[str, ...]] = {
    DagAccessEntity.AUDIT_LOG: (RESOURCE_AUDIT_LOG,),
    DagAccessEntity.CODE: (RESOURCE_DAG_CODE,),
    DagAccessEntity.DEPENDENCIES: (RESOURCE_DAG_DEPENDENCIES,),
    DagAccessEntity.RUN: (RESOURCE_DAG_RUN,),
    # RESOURCE_TASK_INSTANCE has been originally misused. RESOURCE_TASK_INSTANCE referred to task definition
    # AND task instances without making the difference
    # To be backward compatible, we translate DagAccessEntity.TASK_INSTANCE to RESOURCE_TASK_INSTANCE AND
    # RESOURCE_DAG_RUN
    # See https://github.com/apache/airflow/pull/34317#discussion_r1355917769
    DagAccessEntity.TASK: (RESOURCE_TASK_INSTANCE,),
    DagAccessEntity.TASK_INSTANCE: (RESOURCE_DAG_RUN, RESOURCE_TASK_INSTANCE),
    DagAccessEntity.TASK_LOGS: (RESOURCE_TASK_LOG,),
    DagAccessEntity.VERSION: (RESOURCE_DAG_VERSION,),
    DagAccessEntity.WARNING: (RESOURCE_DAG_WARNING,),
    DagAccessEntity.XCOM: (RESOURCE_XCOM,),
}

_MAP_ACCESS_VIEW_TO_FAB_RESOURCE_TYPE = {
    AccessView.CLUSTER_ACTIVITY: RESOURCE_CLUSTER_ACTIVITY,
    AccessView.DOCS: RESOURCE_DOCS,
    AccessView.IMPORT_ERRORS: RESOURCE_IMPORT_ERROR,
    AccessView.JOBS: RESOURCE_JOB,
    AccessView.PLUGINS: RESOURCE_PLUGIN,
    AccessView.PROVIDERS: RESOURCE_PROVIDER,
    AccessView.TRIGGERS: RESOURCE_TRIGGER,
    AccessView.WEBSITE: RESOURCE_WEBSITE,
}

_MAP_MENU_ITEM_TO_FAB_RESOURCE_TYPE = {
    MenuItem.ASSETS: RESOURCE_ASSET,
    MenuItem.AUDIT_LOG: RESOURCE_AUDIT_LOG,
    MenuItem.CONFIG: RESOURCE_CONFIG,
    MenuItem.CONNECTIONS: RESOURCE_CONNECTION,
    MenuItem.DAGS: RESOURCE_DAG,
    MenuItem.DOCS: RESOURCE_DOCS,
    MenuItem.PLUGINS: RESOURCE_PLUGIN,
    MenuItem.POOLS: RESOURCE_POOL,
    MenuItem.PROVIDERS: RESOURCE_PROVIDER,
    MenuItem.VARIABLES: RESOURCE_VARIABLE,
    MenuItem.XCOMS: RESOURCE_XCOM,
}


if AIRFLOW_V_3_1_PLUS:
    from airflow.providers.fab.www.security.permissions import RESOURCE_HITL_DETAIL

    _MAP_MENU_ITEM_TO_FAB_RESOURCE_TYPE[MenuItem.REQUIRED_ACTIONS] = RESOURCE_HITL_DETAIL
    _MAP_DAG_ACCESS_ENTITY_TO_FAB_RESOURCE_TYPE[DagAccessEntity.HITL_DETAIL] = (RESOURCE_HITL_DETAIL,)


class FabAuthManager(BaseAuthManager[User]):
    """
    Flask-AppBuilder auth manager.

    This auth manager is responsible for providing a backward compatible user management experience to users.
    """

    appbuilder: AirflowAppBuilder | None = None

    def init_flask_resources(self) -> None:
        self._sync_appbuilder_roles()

    @cached_property
    def apiserver_endpoint(self) -> str:
        return conf.get("api", "base_url", fallback="/")

    @staticmethod
    def get_cli_commands() -> list[CLICommand]:
        """Vends CLI commands to be included in Airflow CLI."""
        commands: list[CLICommand] = [
            GroupCommand(
                name="users",
                help="Manage users",
                subcommands=USERS_COMMANDS,
            ),
            GroupCommand(
                name="roles",
                help="Manage roles",
                subcommands=ROLES_COMMANDS,
            ),
            SYNC_PERM_COMMAND,  # not in a command group
            PERMISSIONS_CLEANUP_COMMAND,  # single command for permissions cleanup
        ]
        # If Airflow version is 3.0.0 or higher, add the fab-db command group
        if packaging.version.parse(
            packaging.version.parse(airflow_version).base_version
        ) >= packaging.version.parse("3.0.0"):
            commands.append(GroupCommand(name="fab-db", help="Manage FAB", subcommands=DB_COMMANDS))
        return commands

    def get_fastapi_app(self) -> FastAPI | None:
        """Get the FastAPI app."""
        from airflow.providers.fab.auth_manager.api_fastapi.routes.login import (
            login_router,
        )
        from airflow.providers.fab.auth_manager.api_fastapi.routes.roles import roles_router

        flask_app = create_app(enable_plugins=False)

        app = FastAPI(
            title="FAB auth manager API",
            description=(
                "This is FAB auth manager API. This API is only available if the auth manager used in "
                "the Airflow environment is FAB auth manager. "
                "This API provides endpoints to manage users and permissions managed by the FAB auth "
                "manager."
            ),
        )

        # Add the login router to the FastAPI app
        app.include_router(login_router)
        app.include_router(roles_router)

        app.mount("/", WSGIMiddleware(flask_app))

        return app

    def get_api_endpoints(self) -> None | Blueprint:
        folder = Path(__file__).parents[0].resolve()  # this is airflow/auth/managers/fab/
        with folder.joinpath("openapi", "v1-flask-api.yaml").open() as f:
            specification = safe_load(f)
        return FlaskApi(
            specification=specification,
            resolver=_LazyResolver(),
            base_path="/fab/v1",
            options={
                "swagger_ui": SWAGGER_ENABLED,
                "swagger_path": SWAGGER_BUNDLE.__fspath__(),
            },
            strict_validation=True,
            validate_responses=True,
            validator_map={"body": _CustomErrorRequestBodyValidator},
        ).blueprint

    def get_user(self) -> User:
        """
        Return the user associated to the user in session.

        Attempt to find the current user in g.user, as defined by the kerberos authentication backend.
        If no such user is found, return the `current_user` local proxy object, linked to the user session.

        """
        from flask_login import current_user

        # If a user has gone through the Kerberos dance, the kerberos authentication manager
        # has linked it with a User model, stored in g.user, and not the session.
        if current_user.is_anonymous and getattr(g, "user", None) is not None and not g.user.is_anonymous:
            return g.user

        return current_user

    def deserialize_user(self, token: dict[str, Any]) -> User:
        with create_session() as session:
            return session.scalars(select(User).where(User.id == int(token["sub"]))).one()

    def serialize_user(self, user: User) -> dict[str, Any]:
        return {"sub": str(user.id)}

    def is_logged_in(self) -> bool:
        """Return whether the user is logged in."""
        user = self.get_user()
        return (
            self.appbuilder
            and self.appbuilder.app.config.get("AUTH_ROLE_PUBLIC", None)
            or (not user.is_anonymous and user.is_active)
        )

    def is_authorized_configuration(
        self,
        *,
        method: ResourceMethod,
        user: User,
        details: ConfigurationDetails | None = None,
    ) -> bool:
        return self._is_authorized(method=method, resource_type=RESOURCE_CONFIG, user=user)

    def is_authorized_connection(
        self,
        *,
        method: ResourceMethod,
        user: User,
        details: ConnectionDetails | None = None,
    ) -> bool:
        return self._is_authorized(method=method, resource_type=RESOURCE_CONNECTION, user=user)

    def is_authorized_dag(
        self,
        *,
        method: ResourceMethod,
        user: User,
        access_entity: DagAccessEntity | None = None,
        details: DagDetails | None = None,
    ) -> bool:
        """
        Return whether the user is authorized to access the dag.

        There are multiple scenarios:

        1. ``method`` is "GET" and no details are provided which means the user wants to list Dags (or sub entities of Dags).
        2. ``access_entity`` is not provided which means the user wants to access the DAG itself and not a sub
            entity (e.g. Task instances).
        3. ``access_entity`` is provided which means the user wants to access a sub entity of the DAG
            (e.g. DAG runs).

            a. If ``method`` is GET, then check the user has READ permissions on the DAG and the sub entity.
            b. Else, check the user has EDIT permissions on the DAG and ``method`` on the sub entity.

        :param method: The method to authorize.
        :param user: The user performing the action.
        :param access_entity: The dag access entity.
        :param details: The dag details.
        """
        if access_entity:
            # If a sub-Dag entity is specified, check whether the user has access to it
            resource_types = self._get_fab_resource_types(access_entity)
            access_entity_authorized = all(
                self._is_authorized(method=method, resource_type=resource_type, user=user)
                for resource_type in resource_types
            )
            if access_entity == DagAccessEntity.RUN and details and details.id:
                # Check using the deprecated permission prefix "DAG Run" to check whether the user has access to dag runs
                is_authorized_run = self._is_authorized(
                    method=method,
                    resource_type=permissions.resource_name(details.id, RESOURCE_DAG_RUN),
                    user=user,
                )
                if not (is_authorized_run or access_entity_authorized):
                    return False
            elif not access_entity_authorized:
                return False

        if method == "GET" and (not details or not details.id):
            # Scenario 1
            return self._is_authorized_list_dags(user=user)
        if not access_entity:
            # Scenario 2
            return self._is_authorized_dag(method=method, details=details, user=user)
        # Scenario 3
        dag_method: ResourceMethod = "GET" if method == "GET" else "PUT"
        if (details and details.id) and not self._is_authorized_dag(
            method=dag_method, details=details, user=user
        ):
            return False
        return True

    def is_authorized_backfill(
        self,
        *,
        method: ResourceMethod,
        user: User,
        details: BackfillDetails | None = None,
    ) -> bool:
        return self._is_authorized(method=method, resource_type=RESOURCE_BACKFILL, user=user)

    def is_authorized_asset(
        self, *, method: ResourceMethod, user: User, details: AssetDetails | None = None
    ) -> bool:
        return self._is_authorized(method=method, resource_type=RESOURCE_ASSET, user=user)

    def is_authorized_asset_alias(
        self,
        *,
        method: ResourceMethod,
        user: User,
        details: AssetAliasDetails | None = None,
    ) -> bool:
        return self._is_authorized(method=method, resource_type=RESOURCE_ASSET_ALIAS, user=user)

    def is_authorized_pool(
        self, *, method: ResourceMethod, user: User, details: PoolDetails | None = None
    ) -> bool:
        return self._is_authorized(method=method, resource_type=RESOURCE_POOL, user=user)

    def is_authorized_variable(
        self,
        *,
        method: ResourceMethod,
        user: User,
        details: VariableDetails | None = None,
    ) -> bool:
        return self._is_authorized(method=method, resource_type=RESOURCE_VARIABLE, user=user)

    def is_authorized_view(self, *, access_view: AccessView, user: User) -> bool:
        # "Docs" are only links in the menu, there is no page associated
        method: ExtendedResourceMethod = "MENU" if access_view == AccessView.DOCS else "GET"
        return self._is_authorized(
            method=method,
            resource_type=_MAP_ACCESS_VIEW_TO_FAB_RESOURCE_TYPE[access_view],
            user=user,
        )

    def is_authorized_custom_view(
        self, *, method: ResourceMethod | str, resource_name: str, user: User
    ) -> bool:
        fab_action_name = get_fab_action_from_method_map().get(method, method)
        return (fab_action_name, resource_name) in self._get_user_permissions(user)

    def filter_authorized_menu_items(self, menu_items: list[MenuItem], user: User) -> list[MenuItem]:
        return [
            menu_item
            for menu_item in menu_items
            if self._is_authorized(
                method="MENU",
                resource_type=_MAP_MENU_ITEM_TO_FAB_RESOURCE_TYPE.get(menu_item, menu_item.value),
                user=user,
            )
        ]

    @provide_session
    def get_authorized_connections(
        self,
        *,
        user: User,
        method: ResourceMethod = "GET",
        session: Session = NEW_SESSION,
    ) -> set[str]:
        """
        Get connection ids (``conn_id``) the user has access to.

        Fab auth manager does not allow fine-grained access with connections. Thus, return all the connection ids.

        :param user: the user
        :param method: the method to filter on
        :param session: the session
        """
        rows = session.execute(select(Connection.conn_id)).scalars().all()
        return set(rows)

    @provide_session
    def get_authorized_dag_ids(
        self,
        *,
        user: User,
        method: ResourceMethod = "GET",
        session: Session = NEW_SESSION,
    ) -> set[str]:
        if self._is_authorized(method=method, resource_type=RESOURCE_DAG, user=user):
            # If user is authorized to access all DAGs, return all DAGs
            return {dag.dag_id for dag in session.execute(select(DagModel.dag_id))}
        if isinstance(user, AnonymousUser):
            return set()
        user_query = (
            session.scalars(
                select(User)
                .options(
                    joinedload(User.roles)
                    .subqueryload(Role.permissions)
                    .options(joinedload(Permission.action), joinedload(Permission.resource))
                )
                .where(User.id == user.id)
            )
            .unique()
            .one()
        )
        roles = user_query.roles

        map_fab_action_name_to_method_name = get_method_from_fab_action_map()
        resources = set()
        for role in roles:
            for permission in role.permissions:
                action = permission.action.name
                if (
                    action in map_fab_action_name_to_method_name
                    and map_fab_action_name_to_method_name[action] == method
                ):
                    resource = permission.resource.name
                    if resource == permissions.RESOURCE_DAG:
                        return {dag.dag_id for dag in session.execute(select(DagModel.dag_id))}
                    if resource.startswith(permissions.RESOURCE_DAG_PREFIX):
                        resources.add(resource[len(permissions.RESOURCE_DAG_PREFIX) :])
        return set(session.scalars(select(DagModel.dag_id).where(DagModel.dag_id.in_(resources))))

    @provide_session
    def get_authorized_pools(
        self,
        *,
        user: User,
        method: ResourceMethod = "GET",
        session: Session = NEW_SESSION,
    ) -> set[str]:
        """
        Get pools the user has access to.

        Fab auth manager does not allow fine-grained access with pools. Thus, return all the pool names.

        :param user: the user
        :param method: the method to filter on
        :param session: the session
        """
        rows = session.execute(select(Pool.pool)).scalars().all()
        return set(rows)

    @provide_session
    def get_authorized_variables(
        self,
        *,
        user: User,
        method: ResourceMethod = "GET",
        session: Session = NEW_SESSION,
    ) -> set[str]:
        """
        Get variable keys the user has access to.

        Fab auth manager does not allow fine-grained access with variables. Thus, return all the variable keys.

        :param user: the user
        :param method: the method to filter on
        :param session: the session
        """
        rows = session.execute(select(Variable.key)).scalars().all()
        return set(rows)

    @cached_property
    def security_manager(self) -> FabAirflowSecurityManagerOverride:
        """Return the security manager specific to FAB."""
        from airflow.providers.fab.auth_manager.security_manager.override import (
            FabAirflowSecurityManagerOverride,
        )

        if not self.appbuilder:
            raise AirflowException("AppBuilder is not initialized.")

        sm_from_config = current_app.config.get("SECURITY_MANAGER_CLASS")
        if sm_from_config:
            if not issubclass(sm_from_config, FabAirflowSecurityManagerOverride):
                raise AirflowConfigException(
                    """Your CUSTOM_SECURITY_MANAGER must extend FabAirflowSecurityManagerOverride."""
                )
            return sm_from_config(self.appbuilder)

        return FabAirflowSecurityManagerOverride(self.appbuilder)

    def get_url_login(self, **kwargs) -> str:
        """Return the login page url."""
        return urljoin(self.apiserver_endpoint, f"{AUTH_MANAGER_FASTAPI_APP_PREFIX}/login/")

    def get_url_logout(self) -> str | None:
        """Return the logout page url."""
        return urljoin(self.apiserver_endpoint, f"{AUTH_MANAGER_FASTAPI_APP_PREFIX}/logout")

    def register_views(self) -> None:
        self.security_manager.register_views()

    def get_extra_menu_items(self, *, user: User) -> list[ExtraMenuItem]:
        # Contains the list of menu items. ``resource_type`` is the name of the resource in FAB
        # permission model to check whether the user is allowed to see this menu item
        items = [
            {
                "resource_type": "List Users",
                "text": "Users",
                "href": f"{AUTH_MANAGER_FASTAPI_APP_PREFIX}/users/list/",
            },
            {
                "resource_type": "List Roles",
                "text": "Roles",
                "href": f"{AUTH_MANAGER_FASTAPI_APP_PREFIX}/roles/list/",
            },
            {
                "resource_type": "Actions",
                "text": "Actions",
                "href": f"{AUTH_MANAGER_FASTAPI_APP_PREFIX}/actions/list/",
            },
            {
                "resource_type": "Resources",
                "text": "Resources",
                "href": f"{AUTH_MANAGER_FASTAPI_APP_PREFIX}/resources/list/",
            },
            {
                "resource_type": "Permission Pairs",
                "text": "Permissions",
                "href": f"{AUTH_MANAGER_FASTAPI_APP_PREFIX}/permissions/list/",
            },
        ]

        return [
            ExtraMenuItem(text=item["text"], href=item["href"])
            for item in items
            if self._is_authorized(method="MENU", resource_type=item["resource_type"], user=user)
        ]

    @staticmethod
    def get_db_manager() -> str | None:
        return "airflow.providers.fab.auth_manager.models.db.FABDBManager"

    def _is_authorized(
        self,
        *,
        method: ExtendedResourceMethod,
        resource_type: str,
        user: User,
    ) -> bool:
        """
        Return whether the user is authorized to perform a given action.

        :param method: the method to perform
        :param resource_type: the type of resource the user attempts to perform the action on
        :param user: the user performing the action

        :meta private:
        """
        fab_action = self._get_fab_action(method)
        user_permissions = self._get_user_permissions(user)

        return (fab_action, resource_type) in user_permissions

    def _is_authorized_dag(
        self,
        method: ResourceMethod,
        details: DagDetails | None,
        user: User,
    ) -> bool:
        """
        Return whether the user is authorized to perform a given action on a DAG.

        :param method: the method to perform
        :param details: details about the DAG
        :param user: the user performing the action

        :meta private:
        """
        is_global_authorized = self._is_authorized(method=method, resource_type=RESOURCE_DAG, user=user)
        if is_global_authorized:
            return True

        if details and details.id:
            # Check whether the user has permissions to access a specific DAG
            resource_dag_name = permissions.resource_name(details.id, RESOURCE_DAG)
            return self._is_authorized(method=method, resource_type=resource_dag_name, user=user)
        return False

    def _is_authorized_list_dags(self, *, user: User) -> bool:
        """
        Return whether the user is authorized to list Dags.

        :param user: the user performing the action

        :meta private:
        """
        is_global_authorized = self._is_authorized(method="GET", resource_type=RESOURCE_DAG, user=user)
        if is_global_authorized:
            return True

        user_permissions = self._get_user_permissions(user)
        for action, resource in user_permissions:
            if action == ACTION_CAN_READ and resource.startswith(RESOURCE_DAG_PREFIX):
                return True
        return False

    @staticmethod
    def _get_fab_action(method: ExtendedResourceMethod) -> str:
        """
        Convert the method to a FAB action.

        :param method: the method to convert

        :meta private:
        """
        fab_action_from_method_map = get_fab_action_from_method_map()
        if method not in fab_action_from_method_map:
            raise AirflowException(f"Unknown method: {method}")
        return fab_action_from_method_map[method]

    @staticmethod
    def _get_fab_resource_types(dag_access_entity: DagAccessEntity) -> tuple[str, ...]:
        """
        Convert a DAG access entity to a tuple of FAB resource type.

        :param dag_access_entity: the DAG access entity

        :meta private:
        """
        if dag_access_entity not in _MAP_DAG_ACCESS_ENTITY_TO_FAB_RESOURCE_TYPE:
            raise AirflowException(f"Unknown DAG access entity: {dag_access_entity}")
        return _MAP_DAG_ACCESS_ENTITY_TO_FAB_RESOURCE_TYPE[dag_access_entity]

    @staticmethod
    def _get_user_permissions(user: User):
        """
        Return the user permissions.

        :param user: the user to get permissions for

        :meta private:
        """
        # If the user gets deleted while being logged in
        if not user:
            return []
        return getattr(user, "perms") or []

    def _sync_appbuilder_roles(self):
        """
        Sync appbuilder roles to DB.

        :meta private:
        """
        # Garbage collect old permissions/views after they have been modified.
        # Otherwise, when the name of a view or menu is changed, the framework
        # will add the new Views and Menus names to the backend, but will not
        # delete the old ones.
        if conf.getboolean("fab", "UPDATE_FAB_PERMS"):
            self.security_manager.sync_roles()


def get_parser() -> argparse.ArgumentParser:
    """Generate documentation; used by Sphinx argparse."""
    from airflow.cli.cli_parser import AirflowHelpFormatter, _add_command

    parser = DefaultHelpParser(prog="airflow", formatter_class=AirflowHelpFormatter)
    subparsers = parser.add_subparsers(dest="subcommand", metavar="GROUP_OR_COMMAND")
    for group_command in FabAuthManager.get_cli_commands():
        _add_command(subparsers, group_command)
    return parser



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, cast

from flask_appbuilder.const import AUTH_LDAP
from starlette import status
from starlette.exceptions import HTTPException

from airflow.api_fastapi.app import get_auth_manager
from airflow.configuration import conf
from airflow.providers.fab.auth_manager.api_fastapi.datamodels.login import LoginBody, LoginResponse

if TYPE_CHECKING:
    from airflow.providers.fab.auth_manager.fab_auth_manager import FabAuthManager
    from airflow.providers.fab.auth_manager.models import User


class FABAuthManagerLogin:
    """Login Service for FABAuthManager."""

    @classmethod
    def create_token(
        cls, body: LoginBody, expiration_time_in_seconds: int = conf.getint("api_auth", "jwt_expiration_time")
    ) -> LoginResponse:
        """Create a new token."""
        if not body.username or not body.password:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST, detail="Username and password must be provided"
            )

        auth_manager = cast("FabAuthManager", get_auth_manager())
        user: User | None = None

        if auth_manager.security_manager.auth_type == AUTH_LDAP:
            user = auth_manager.security_manager.auth_user_ldap(
                body.username, body.password, rotate_session_id=False
            )
        if user is None:
            user = auth_manager.security_manager.auth_user_db(
                body.username, body.password, rotate_session_id=False
            )

        if not user:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid credentials")

        return LoginResponse(
            access_token=auth_manager.generate_jwt(
                user=user, expiration_time_in_seconds=expiration_time_in_seconds
            )
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow import settings
from airflow.cli.commands.db_command import run_db_downgrade_command, run_db_migrate_command
from airflow.providers.fab.auth_manager.models.db import _REVISION_HEADS_MAP, FABDBManager
from airflow.utils import cli as cli_utils
from airflow.utils.providers_configuration_loader import providers_configuration_loaded


@providers_configuration_loaded
def resetdb(args):
    """Reset the metadata database."""
    print(f"DB: {settings.engine.url!r}")
    if not (args.yes or input("This will drop existing tables if they exist. Proceed? (y/n)").upper() == "Y"):
        raise SystemExit("Cancelled")
    FABDBManager(settings.Session()).resetdb(skip_init=args.skip_init)


@cli_utils.action_cli(check_db=False)
@providers_configuration_loaded
def migratedb(args):
    """Migrates the metadata database."""
    session = settings.Session()
    upgrade_command = FABDBManager(session).upgradedb
    run_db_migrate_command(args, upgrade_command, revision_heads_map=_REVISION_HEADS_MAP)


@cli_utils.action_cli(check_db=False)
@providers_configuration_loaded
def downgrade(args):
    """Downgrades the metadata database."""
    session = settings.Session()
    dwongrade_command = FABDBManager(session).downgrade
    run_db_downgrade_command(args, dwongrade_command, revision_heads_map=_REVISION_HEADS_MAP)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Test permissions command."""

from __future__ import annotations

import argparse
from contextlib import redirect_stdout
from importlib import reload
from io import StringIO
from unittest.mock import MagicMock, patch

import pytest

from airflow.cli import cli_parser
from airflow.providers.fab.auth_manager.cli_commands import permissions_command
from airflow.providers.fab.auth_manager.cli_commands.utils import get_application_builder

from tests_common.test_utils.config import conf_vars

pytestmark = pytest.mark.db_test


class TestPermissionsCommand:
    """Test permissions cleanup CLI commands."""

    @pytest.fixture(autouse=True)
    def _set_attrs(self):
        with conf_vars(
            {
                (
                    "core",
                    "auth_manager",
                ): "airflow.providers.fab.auth_manager.fab_auth_manager.FabAuthManager",
            }
        ):
            # Reload the module to use FAB auth manager
            reload(cli_parser)
            # Clearing the cache before calling it
            cli_parser.get_parser.cache_clear()
            self.parser = cli_parser.get_parser()
            with get_application_builder() as appbuilder:
                self.appbuilder = appbuilder
                yield

    @patch("airflow.providers.fab.auth_manager.cli_commands.permissions_command.cleanup_dag_permissions")
    @patch("airflow.providers.fab.auth_manager.models.Resource")
    def test_permissions_cleanup_success(self, mock_resource, mock_cleanup_dag_permissions):
        """Test successful cleanup of DAG permissions."""
        # Mock args
        args = argparse.Namespace()
        args.dag_id = None
        args.dry_run = False
        args.yes = True
        args.verbose = True

        # Mock orphaned resources
        mock_orphaned_resource = MagicMock()
        mock_orphaned_resource.name = "DAG:orphaned_dag"

        with (
            patch("airflow.providers.fab.auth_manager.cli_commands.utils.get_application_builder"),
            patch("airflow.utils.session.create_session") as mock_session_ctx,
            patch("sqlalchemy.select"),
            redirect_stdout(StringIO()),
        ):
            mock_session = MagicMock()
            mock_session_ctx.return_value.__enter__.return_value = mock_session

            # Mock DagModel query - return existing DAGs
            mock_dag_result = MagicMock()
            mock_dag_result.all.return_value = [MagicMock(dag_id="existing_dag")]

            # Mock Resource query - return orphaned resources
            mock_resource_result = MagicMock()
            mock_resource_result.all.return_value = [mock_orphaned_resource]

            # Setup session.scalars to return different results for different queries
            mock_session.scalars.side_effect = [mock_dag_result, mock_resource_result]

            permissions_command.permissions_cleanup(args)

        # Verify function calls - it should be called exactly once for the orphaned DAG
        mock_cleanup_dag_permissions.assert_called_once_with("orphaned_dag", mock_session)

    @patch("airflow.providers.fab.auth_manager.cli_commands.permissions_command.cleanup_dag_permissions")
    @patch("airflow.providers.fab.auth_manager.models.Resource")
    def test_permissions_cleanup_dry_run(self, mock_resource, mock_cleanup_dag_permissions):
        """Test dry run mode for permissions cleanup."""
        # Mock args
        args = argparse.Namespace()
        args.dag_id = None
        args.dry_run = True
        args.verbose = True

        # Mock orphaned resources
        mock_orphaned_resource = MagicMock()
        mock_orphaned_resource.name = "DAG:orphaned_dag"

        with (
            patch("airflow.providers.fab.auth_manager.cli_commands.utils.get_application_builder"),
            patch("airflow.utils.session.create_session") as mock_session_ctx,
            patch("sqlalchemy.select"),
            redirect_stdout(StringIO()) as stdout,
        ):
            mock_session = MagicMock()
            mock_session_ctx.return_value.__enter__.return_value = mock_session

            # Mock DagModel query - return existing DAGs
            mock_dag_result = MagicMock()
            mock_dag_result.all.return_value = [MagicMock(dag_id="existing_dag")]

            # Mock Resource query - return orphaned resources
            mock_resource_result = MagicMock()
            mock_resource_result.all.return_value = [mock_orphaned_resource]

            # Setup session.scalars to return different results for different queries
            mock_session.scalars.side_effect = [mock_dag_result, mock_resource_result]

            permissions_command.permissions_cleanup(args)

        output = stdout.getvalue()
        assert "Dry run mode" in output or "No orphaned DAG permissions found" in output
        # In dry run mode, cleanup_dag_permissions should NOT be called
        mock_cleanup_dag_permissions.assert_not_called()

    @patch("airflow.providers.fab.auth_manager.cli_commands.permissions_command.cleanup_dag_permissions")
    @patch("airflow.providers.fab.auth_manager.models.Resource")
    def test_permissions_cleanup_specific_dag(self, mock_resource, mock_cleanup_dag_permissions):
        """Test cleanup for a specific DAG."""
        # Mock args
        args = argparse.Namespace()
        args.dag_id = "test_dag"
        args.dry_run = False
        args.yes = True
        args.verbose = True

        # Mock orphaned resource for the specific DAG
        mock_orphaned_resource = MagicMock()
        mock_orphaned_resource.name = "DAG:test_dag"

        with (
            patch("airflow.providers.fab.auth_manager.cli_commands.utils.get_application_builder"),
            patch("airflow.utils.session.create_session") as mock_session_ctx,
            patch("sqlalchemy.select"),
            redirect_stdout(StringIO()),
        ):
            mock_session = MagicMock()
            mock_session_ctx.return_value.__enter__.return_value = mock_session

            # Mock DagModel query - return existing DAGs (NOT including the target DAG)
            mock_dag_result = MagicMock()
            mock_dag_result.all.return_value = [
                MagicMock(dag_id="existing_dag"),
                MagicMock(dag_id="another_existing_dag"),
            ]

            # Mock Resource query - return orphaned resources
            mock_resource_result = MagicMock()
            mock_resource_result.all.return_value = [mock_orphaned_resource]

            # Setup session.scalars to return different results for different queries
            mock_session.scalars.side_effect = [mock_dag_result, mock_resource_result]

            permissions_command.permissions_cleanup(args)

        # Should call cleanup_dag_permissions specifically for test_dag
        mock_cleanup_dag_permissions.assert_called_once_with("test_dag", mock_session)

    @patch("airflow.providers.fab.auth_manager.cli_commands.permissions_command.cleanup_dag_permissions")
    @patch("airflow.providers.fab.auth_manager.models.Resource")
    @patch("builtins.input", return_value="n")
    def test_permissions_cleanup_no_confirmation(
        self, mock_input, mock_resource, mock_cleanup_dag_permissions
    ):
        """Test cleanup cancellation when user doesn't confirm."""
        # Mock args
        args = argparse.Namespace()
        args.dag_id = None
        args.dry_run = False
        args.yes = False
        args.verbose = False

        # Mock orphaned resources
        mock_orphaned_resource = MagicMock()
        mock_orphaned_resource.name = "DAG:orphaned_dag"

        with (
            patch("airflow.providers.fab.auth_manager.cli_commands.utils.get_application_builder"),
            patch("airflow.utils.session.create_session") as mock_session_ctx,
            patch("sqlalchemy.select"),
            redirect_stdout(StringIO()) as stdout,
        ):
            mock_session = MagicMock()
            mock_session_ctx.return_value.__enter__.return_value = mock_session

            # Mock DagModel query - return existing DAGs
            mock_dag_result = MagicMock()
            mock_dag_result.all.return_value = [MagicMock(dag_id="existing_dag")]

            # Mock Resource query - return orphaned resources
            mock_resource_result = MagicMock()
            mock_resource_result.all.return_value = [mock_orphaned_resource]

            # Setup session.scalars to return different results for different queries
            mock_session.scalars.side_effect = [mock_dag_result, mock_resource_result]

            permissions_command.permissions_cleanup(args)

        output = stdout.getvalue()
        # Should not call cleanup if user declines or no orphaned permissions found
        assert "Cleanup cancelled" in output or "No orphaned DAG permissions found" in output

        # cleanup_dag_permissions should NOT be called when user cancels
        if "Cleanup cancelled" in output:
            mock_cleanup_dag_permissions.assert_not_called()


class TestDagPermissions:
    """Test cases for cleanup_dag_permissions function with real database operations."""

    @pytest.fixture(autouse=True)
    def _setup_fab_test(self):
        """Setup FAB for testing."""
        with conf_vars(
            {
                (
                    "core",
                    "auth_manager",
                ): "airflow.providers.fab.auth_manager.fab_auth_manager.FabAuthManager",
            }
        ):
            with get_application_builder():
                yield

    def test_cleanup_dag_permissions_removes_specific_dag_resources(self):
        """Test that cleanup_dag_permissions removes only the specified DAG resources."""
        from sqlalchemy import select

        from airflow.providers.fab.auth_manager.cli_commands.permissions_command import (
            cleanup_dag_permissions,
        )
        from airflow.providers.fab.auth_manager.models import Action, Permission, Resource
        from airflow.security.permissions import RESOURCE_DAG_PREFIX
        from airflow.utils.session import create_session

        with create_session() as session:
            # Create resources for two different DAGs
            target_resource = Resource(name=f"{RESOURCE_DAG_PREFIX}target_dag")
            keep_resource = Resource(name=f"{RESOURCE_DAG_PREFIX}keep_dag")
            session.add_all([target_resource, keep_resource])
            session.flush()

            # Get or create action
            read_action = session.scalars(select(Action).where(Action.name == "can_read")).first()
            if not read_action:
                read_action = Action(name="can_read")
                session.add(read_action)
                session.flush()

            # Create permissions
            target_perm = Permission(action=read_action, resource=target_resource)
            keep_perm = Permission(action=read_action, resource=keep_resource)
            session.add_all([target_perm, keep_perm])
            session.commit()

            # Execute cleanup
            cleanup_dag_permissions("target_dag", session)

            # Verify: target resource deleted, keep resource remains
            assert not session.get(Resource, target_resource.id)
            assert session.get(Resource, keep_resource.id)
            assert not session.get(Permission, target_perm.id)
            assert session.get(Permission, keep_perm.id)

    def test_cleanup_dag_permissions_handles_no_matching_resources(self):
        """Test that cleanup_dag_permissions handles DAGs with no matching resources gracefully."""
        from sqlalchemy import func, select

        from airflow.providers.fab.auth_manager.cli_commands.permissions_command import (
            cleanup_dag_permissions,
        )
        from airflow.providers.fab.auth_manager.models import Resource
        from airflow.utils.session import create_session

        with create_session() as session:
            initial_count = session.scalar(select(func.count(Resource.id)))
            cleanup_dag_permissions("non_existent_dag", session)
            assert session.scalar(select(func.count(Resource.id))) == initial_count

    def test_cleanup_dag_permissions_handles_resources_without_permissions(self):
        """Test cleanup when resources exist but have no permissions."""
        from airflow.providers.fab.auth_manager.cli_commands.permissions_command import (
            cleanup_dag_permissions,
        )
        from airflow.providers.fab.auth_manager.models import Resource
        from airflow.security.permissions import RESOURCE_DAG_PREFIX
        from airflow.utils.session import create_session

        with create_session() as session:
            # Create resource without permissions
            resource = Resource(name=f"{RESOURCE_DAG_PREFIX}test_dag")
            session.add(resource)
            session.commit()
            resource_id = resource.id

            cleanup_dag_permissions("test_dag", session)
            assert not session.get(Resource, resource_id)

    def test_cleanup_dag_permissions_with_default_session(self):
        """Test cleanup_dag_permissions when no session is provided (uses default)."""
        from sqlalchemy import func, select

        from airflow.providers.fab.auth_manager.cli_commands.permissions_command import (
            cleanup_dag_permissions,
        )
        from airflow.providers.fab.auth_manager.models import Resource
        from airflow.security.permissions import RESOURCE_DAG_PREFIX
        from airflow.utils.session import create_session

        # Setup test data
        with create_session() as session:
            resource = Resource(name=f"{RESOURCE_DAG_PREFIX}test_dag")
            session.add(resource)
            session.commit()

        # Call cleanup without session parameter
        cleanup_dag_permissions("test_dag")

        # Verify deletion
        with create_session() as session:
            count = session.scalar(
                select(func.count(Resource.id)).where(Resource.name == f"{RESOURCE_DAG_PREFIX}test_dag")
            )
            assert count == 0



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from contextlib import nullcontext as _noop_cm
from unittest.mock import ANY, MagicMock, patch

import pytest
from fastapi import HTTPException, status

from airflow.providers.fab.auth_manager.api_fastapi.datamodels.roles import (
    RoleCollectionResponse,
    RoleResponse,
)


@pytest.mark.db_test
class TestRoles:
    # POST /roles

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_create_role(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        dummy_out = RoleResponse(name="my_new_role", permissions=[])
        mock_roles.create_role.return_value = dummy_out

        with as_user():
            resp = test_client.post("/fab/v1/roles", json={"name": "my_new_role", "actions": []})
            assert resp.status_code == 200
            assert resp.json() == dummy_out.model_dump(by_alias=True)
            mock_roles.create_role.assert_called_once()

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_create_role_forbidden(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = False
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.post("/fab/v1/roles", json={"name": "r", "actions": []})
            assert resp.status_code == 403
            mock_roles.create_role.assert_not_called()

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_create_role_validation_422_empty_name(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.post("/fab/v1/roles", json={"name": "", "actions": []})
            assert resp.status_code == 422
            mock_roles.create_role.assert_not_called()

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_create_role_validation_422_missing_name(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.post("/fab/v1/roles", json={"actions": []})
            assert resp.status_code == 422
            mock_roles.create_role.assert_not_called()

    # GET /roles

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    @patch("airflow.providers.fab.auth_manager.api_fastapi.parameters.conf")
    def test_get_roles_success_defaults(
        self,
        conf_mock,
        mock_get_application_builder,
        mock_get_auth_manager,
        mock_roles,
        test_client,
        as_user,
    ):
        conf_mock.getint.side_effect = lambda section, option: {
            "maximum_page_limit": 500,
            "fallback_page_limit": 25,
        }[option]

        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        dummy = RoleCollectionResponse(
            roles=[RoleResponse(name="viewer", permissions=[])],
            total_entries=1,
            limit=100,
            offset=0,
        )
        mock_roles.get_roles.return_value = dummy

        with as_user():
            resp = test_client.get("/fab/v1/roles")
            assert resp.status_code == 200
            assert resp.json() == dummy.model_dump(by_alias=True)
            mock_roles.get_roles.assert_called_once_with(order_by="name", limit=100, offset=0)

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    @patch("airflow.providers.fab.auth_manager.api_fastapi.parameters.conf")
    def test_get_roles_passes_params_and_clamps_limit(
        self,
        conf_mock,
        mock_get_application_builder,
        mock_get_auth_manager,
        mock_roles,
        test_client,
        as_user,
    ):
        conf_mock.getint.side_effect = lambda section, option: {
            "maximum_page_limit": 50,
            "fallback_page_limit": 20,
        }[option]

        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        dummy = RoleCollectionResponse(roles=[], total_entries=0, limit=50, offset=7)
        mock_roles.get_roles.return_value = dummy

        with as_user():
            resp = test_client.get("/fab/v1/roles", params={"order_by": "-name", "limit": 1000, "offset": 7})
            assert resp.status_code == 200
            assert resp.json() == dummy.model_dump(by_alias=True)
            mock_roles.get_roles.assert_called_once_with(order_by="-name", limit=50, offset=7)

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    @patch("airflow.providers.fab.auth_manager.api_fastapi.parameters.conf")
    def test_get_roles_uses_fallback_when_limit_zero(
        self,
        conf_mock,
        mock_get_application_builder,
        mock_get_auth_manager,
        mock_roles,
        test_client,
        as_user,
    ):
        conf_mock.getint.side_effect = lambda section, option: {
            "maximum_page_limit": 100,
            "fallback_page_limit": 33,
        }[option]

        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        dummy = RoleCollectionResponse(roles=[], total_entries=0, limit=33, offset=0)
        mock_roles.get_roles.return_value = dummy

        with as_user():
            resp = test_client.get("/fab/v1/roles", params={"limit": 0})
            assert resp.status_code == 200
            assert resp.json() == dummy.model_dump(by_alias=True)
            mock_roles.get_roles.assert_called_once_with(order_by="name", limit=33, offset=0)

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_get_roles_forbidden(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = False
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.get("/fab/v1/roles")
            assert resp.status_code == 403
            mock_roles.get_roles.assert_not_called()

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_get_roles_validation_422_negative_offset(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.get("/fab/v1/roles", params={"offset": -1})
            assert resp.status_code == 422
            mock_roles.get_roles.assert_not_called()

    # DELETE /roles/{name}

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_delete_role(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_roles.delete_role.return_value = None
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.delete("/fab/v1/roles/roleA")
            assert resp.status_code == 200
            mock_roles.delete_role.assert_called_once_with(name="roleA")

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_delete_role_forbidden(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = False
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.delete("/fab/v1/roles/roleA")
            assert resp.status_code == 403
            mock_roles.delete_role.assert_not_called()

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_delete_role_validation_404_not_found(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr
        mock_roles.delete_role.side_effect = HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Role with name 'non_existent_role' does not exist.",
        )

        with as_user():
            resp = test_client.delete("/fab/v1/roles/non_existent_role")
            assert resp.status_code == 404
            mock_roles.delete_role.assert_called_once_with(name="non_existent_role")

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_delete_role_validation_404_empty_name(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.delete("/fab/v1/roles/")
            assert resp.status_code == 404
            mock_roles.delete_role.assert_not_called()

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_get_role(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        dummy_out = RoleResponse(name="roleA", permissions=[])
        mock_roles.get_role.return_value = dummy_out

        with as_user():
            resp = test_client.get("/fab/v1/roles/roleA")
            assert resp.status_code == 200
            assert resp.json() == dummy_out.model_dump(by_alias=True)
            mock_roles.get_role.assert_called_once_with(name="roleA")

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_get_role_forbidden(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = False
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.get("/fab/v1/roles/roleA")
            assert resp.status_code == 403
            mock_roles.get_role.assert_not_called()

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_get_role_validation_404_not_found(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr
        mock_roles.get_role.side_effect = HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Role with name 'non_existent_role' does not exist.",
        )

        with as_user():
            resp = test_client.get("/fab/v1/roles/non_existent_role")
            assert resp.status_code == 404
            mock_roles.get_role.assert_called_once_with(name="non_existent_role")

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_get_role_validation_404_empty_name(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.get("/fab/v1/roles/")
            assert resp.status_code == 404
            mock_roles.get_role.assert_not_called()

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_patch_role(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        dummy_out = RoleResponse(name="roleA", permissions=[])
        mock_roles.patch_role.return_value = dummy_out

        with as_user():
            resp = test_client.patch("/fab/v1/roles/roleA", json={"name": "roleA", "actions": []})
            assert resp.status_code == 200
            assert resp.json() == dummy_out.model_dump(by_alias=True)
            mock_roles.patch_role.assert_called_once_with(name="roleA", body=ANY, update_mask=None)

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_patch_role_with_update_mask(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        dummy_out = RoleResponse(name="roleA", permissions=[])
        mock_roles.patch_role.return_value = dummy_out

        with as_user():
            resp = test_client.patch(
                "/fab/v1/roles/roleA",
                json={"name": "roleA", "actions": []},
                params={"update_mask": "name,actions"},
            )
            assert resp.status_code == 200
            assert resp.json() == dummy_out.model_dump(by_alias=True)
            mock_roles.patch_role.assert_called_once_with(name="roleA", body=ANY, update_mask="name,actions")

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_path_role_forbidden(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = False
        mock_get_auth_manager.return_value = mgr

        with as_user():
            resp = test_client.patch("/fab/v1/roles/roleA", json={"name": "roleA", "actions": []})
            assert resp.status_code == 403
            mock_roles.patch_role.assert_not_called()

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_patch_role_validation_404_not_found(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr
        mock_roles.patch_role.side_effect = HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Role with name 'non_existent_role' does not exist.",
        )

        with as_user():
            resp = test_client.patch(
                "/fab/v1/roles/non_existent_role", json={"name": "non_existent_role", "actions": []}
            )
            assert resp.status_code == 404
            mock_roles.patch_role.assert_called_once_with(
                name="non_existent_role", body=ANY, update_mask=None
            )

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_patch_role_validation_404_empty_name(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr
        mock_roles.patch_role.side_effect = HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Role with name 'non_existent_role' does not exist.",
        )

        with as_user():
            resp = test_client.patch("/fab/v1/roles/", json={"name": "non_existent_role", "actions": []})
            assert resp.status_code == 404

    @patch("airflow.providers.fab.auth_manager.api_fastapi.routes.roles.FABAuthManagerRoles")
    @patch("airflow.providers.fab.auth_manager.api_fastapi.security.get_auth_manager")
    @patch(
        "airflow.providers.fab.auth_manager.api_fastapi.routes.roles.get_application_builder",
        return_value=_noop_cm(),
    )
    def test_path_role_unknown_update_mask(
        self, mock_get_application_builder, mock_get_auth_manager, mock_roles, test_client, as_user
    ):
        mgr = MagicMock()
        mgr.is_authorized_custom_view.return_value = True
        mock_get_auth_manager.return_value = mgr

        mock_roles.patch_role.side_effect = HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="unknown_field in update_mask is unknown",
        )

        with as_user():
            resp = test_client.patch(
                "/fab/v1/roles/roleA",
                json={"name": "roleA", "actions": []},
                params={"update_mask": "unknown_field"},
            )
            assert resp.status_code == 400
            mock_roles.patch_role.assert_called_once_with(name="roleA", body=ANY, update_mask="unknown_field")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import pytest
from fastapi import HTTPException
from sqlalchemy import column, func
from sqlalchemy.dialects import postgresql
from sqlalchemy.sql.elements import ColumnElement

from airflow.providers.fab.auth_manager.api_fastapi.sorting import build_ordering


def _sql(expr) -> str:
    return str(expr.compile(dialect=postgresql.dialect()))


def test_build_ordering_returns_asc_by_default():
    allowed = {"name": column("name"), "id": column("id")}
    expr = build_ordering("name", allowed=allowed)

    assert isinstance(expr, ColumnElement)
    sql = _sql(expr)
    assert "name" in sql
    assert "ASC" in sql
    assert "DESC" not in sql


def test_build_ordering_desc_when_prefixed_with_dash():
    allowed = {"name": column("name"), "id": column("id")}
    expr = build_ordering("-id", allowed=allowed)

    sql = _sql(expr)
    assert "id" in sql
    assert "DESC" in sql


def test_build_ordering_supports_sql_expressions():
    # Ensure mapping can include arbitrary SQL expressions (e.g., lower(name))
    allowed = {"name_i": func.lower(column("name"))}
    expr = build_ordering("-name_i", allowed=allowed)

    sql = _sql(expr)
    assert "lower(" in sql
    assert "DESC" in sql


def test_build_ordering_raises_http_400_for_disallowed_key():
    allowed = {"name": column("name")}
    with pytest.raises(HTTPException) as ex:
        build_ordering("unknown", allowed=allowed)

    assert ex.value.status_code == 400
    assert "disallowed" in str(ex.value.detail) or "does not exist" in str(ex.value.detail)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import types

import pytest
from pydantic import ValidationError

from airflow.providers.fab.auth_manager.api_fastapi.datamodels.roles import (
    ActionResourceResponse,
    ActionResponse,
    ResourceResponse,
    RoleBody,
    RoleCollectionResponse,
    RoleResponse,
)


class TestRoleModels:
    def test_rolebody_accepts_actions_alias_and_maps_to_permissions(self):
        data = {
            "name": "viewer",
            "actions": [
                {"action": {"name": "can_read"}, "resource": {"name": "DAG"}},
                {"action": {"name": "can_read"}, "resource": {"name": "Connection"}},
            ],
        }
        body = RoleBody.model_validate(data)
        assert body.name == "viewer"
        # Field(validation_alias="actions") should populate `permissions`
        assert len(body.permissions) == 2
        assert body.permissions[0].action.name == "can_read"
        assert body.permissions[0].resource.name == "DAG"

    def test_rolebody_defaults_permissions_to_empty_when_actions_missing(self):
        body = RoleBody.model_validate({"name": "empty"})
        assert body.name == "empty"
        assert body.permissions == []

    def test_rolebody_name_min_length_enforced(self):
        with pytest.raises(ValidationError):
            RoleBody.model_validate({"name": "", "actions": []})

    def test_roleresponse_serializes_permissions_under_actions_alias(self):
        ar = ActionResourceResponse(
            action=ActionResponse(name="can_read"),
            resource=ResourceResponse(name="DAG"),
        )
        rr = RoleResponse(name="viewer", permissions=[ar])

        dumped = rr.model_dump(by_alias=True)
        # Field(serialization_alias="actions") should rename `permissions` -> `actions`
        assert "actions" in dumped
        assert "permissions" not in dumped
        assert dumped["name"] == "viewer"
        assert dumped["actions"][0]["action"]["name"] == "can_read"
        assert dumped["actions"][0]["resource"]["name"] == "DAG"

    def test_roleresponse_model_validate_from_simple_namespace(self):
        # Service returns plain objects; ensure model_validate handles them
        obj = types.SimpleNamespace(
            name="viewer",
            permissions=[
                types.SimpleNamespace(
                    action=types.SimpleNamespace(name="can_read"),
                    resource=types.SimpleNamespace(name="DAG"),
                )
            ],
        )
        rr = RoleResponse.model_validate(obj)
        assert rr.name == "viewer"
        assert rr.permissions
        first = rr.permissions[0]
        assert first.action.name == "can_read"

    def test_rolecollection_response_dump_and_counts(self):
        ar = ActionResourceResponse(
            action=ActionResponse(name="can_read"),
            resource=ResourceResponse(name="DAG"),
        )
        rc = RoleCollectionResponse(
            roles=[RoleResponse(name="viewer", permissions=[ar])],
            total_entries=1,
        )
        dumped = rc.model_dump(by_alias=True)
        assert dumped["total_entries"] == 1
        assert isinstance(dumped["roles"], list)
        assert dumped["roles"][0]["name"] == "viewer"
        assert "actions" in dumped["roles"][0]
        assert "permissions" not in dumped["roles"][0]

    def test_rolecollection_model_validate_from_objects(self):
        obj = types.SimpleNamespace(
            roles=[
                types.SimpleNamespace(
                    name="admin",
                    permissions=[
                        types.SimpleNamespace(
                            action=types.SimpleNamespace(name="can_read"),
                            resource=types.SimpleNamespace(name="DAG"),
                        )
                    ],
                )
            ],
            total_entries=1,
        )
        rc = RoleCollectionResponse.model_validate(obj)
        assert rc.total_entries == 1
        assert len(rc.roles) == 1
        assert rc.roles[0].name == "admin"
        assert rc.roles[0].permissions[0].action.name == "can_read"

    def test_rolecollection_missing_total_entries_raises(self):
        with pytest.raises(ValidationError):
            RoleCollectionResponse.model_validate({"roles": []})
